------------------------------------------------------------------------------
--                                                                          --
--                          GNATCHECK COMPONENTS                            --
--                                                                          --
--             G N A T C H E C K . G L O B A L _ S T A T E . C G            --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2005-2006, AdaCore                     --
--                                                                          --
-- GNATCHECK  is  free  software;  you can redistribute it and/or modify it --
-- under terms of the  GNU  General Public License as published by the Free --
-- Software Foundation;  either version 2, or ( at your option)  any  later --
-- version.  GNATCHECK  is  distributed in the hope that it will be useful, --
-- but  WITHOUT  ANY  WARRANTY;   without  even  the  implied  warranty  of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details.  You should have received a copy of the --
-- GNU  General Public License distributed with GNAT; see file  COPYING. If --
-- not,  write to the  Free Software Foundation,  51 Franklin Street, Fifth --
-- Floor, Boston, MA 02110-1301, USA.                                       --
--                                                                          --
-- GNATCHECK is maintained by AdaCore (http://www.adacore.com).             --
--                                                                          --
------------------------------------------------------------------------------

with Asis.Declarations;          use Asis.Declarations;
with Asis.Definitions;           use Asis.Definitions;
with Asis.Elements;              use Asis.Elements;
with Asis.Expressions;           use Asis.Expressions;
with Asis.Extensions;            use Asis.Extensions;
with Asis.Extensions.Flat_Kinds; use Asis.Extensions.Flat_Kinds;
with Asis.Statements;            use Asis.Statements;

with ASIS_UL.Output;             use ASIS_UL.Output;

with Gnatcheck.ASIS_Utilities; use Gnatcheck.ASIS_Utilities;
with Gnatcheck.Rules.Traversing;

package body Gnatcheck.Global_State.CG is

   ------------------------
   --  Local subprograms --
   ------------------------

   procedure Check_CG_Completeness;
   --  Checks if the information stored in the global data structure is
   --  complete and allows to construct the full Call Graph. Generates a
   --  diagnostic message each time when some incompleteness is detected.

   procedure Traverse_Renamings;
   --  This procedure goes trough all the Call Graph nodes and sets or updates
   --  (direct) call/caller chains for subprogram renaming entities and for
   --  subprogram entities that have renaming-as-body as completion. See the
   --  body of Gnatcheck.Global_State for full documentation of representing
   --  renamiings in global structure

   procedure Process_Call (Call : Asis.Element);
   --  For the argument that should be Gnatcheck.Asis_Utilities.Is_Call,
   --  analyzes the call, tries to locate a called entity and creates the
   --  corresponding arc in the call graph. Each arc is stored only once. If
   --  the called entity cannot  be defined, creates a special arc that
   --  describes why the called entity cannot  be located.

   procedure Process_Task_Entry_Call (Call : Asis.Element);
   procedure Process_Protected_Operation_Call (Call : Asis.Element);
   procedure Process_Ordinary_Call (Call : Asis.Element);
   --  Each of these three procedures implements the corresponding part of
   --  the functionality of Process_Call

   procedure Process_Elab_Calls
     (El      : Asis.Element;
      Control : in out Traverse_Control;
      State   : in out Rule_Traversal_State);
   --  For the argument that should be
   --  Gnatcheck.Asis_Utilities.May_Contain_Elab_Calls, analyzes the argument,
   --  tries to find implicit calls that are made during the elaboration and
   --  for each of these calls processes this call as a regular call.

   procedure Process_Renaming (El : Asis.Element);
   --  Provided that Is_Renaming (El), analyzed El and creates the
   --  corresponding structures in the call graph.
   --  More details???
   --  See the body of Gnatcheck.Global_State for full documentation of
   --  representing renamiings in global structure

   procedure Store_Dispatching_Call;
   --  Stores for the current scope the arc representing the dispatching call.
   --  This arc is stored only once for every caller node.
   --  Is this the right place for declaring this procedure?

   procedure Store_Dynamic_Call;
   --  Stores for the current scope the arc representing the dynamic call.
   --  This arc is stored only once for every caller node.
   --  Is this the right place for declaring this procedure?

   procedure Closure_Step (Remains_Unclosed_Node : out Boolean);
   --  Performs one closure step. See the body for details.
   --  ??? More explanations for the spec?
   --  Remains_Unclosed_Node is set to True if after performing this step there
   --  still may be some node that needs further processing, and to False
   --  otherwise.

   procedure Analyze_Node (N : GS_Node_Id; Status_Changed : out Boolean);
   --  This is a part of Closure_Step, performed for one node, for that at
   --  least one of two flags All_Calls_Set or All_Callers_Set should be still
   --  OFF. It checks if the list of called (calling) entities is either empty
   --  or for all the elements in this list All_Calls_Set (All_Callers_Set) are
   --  already ON (that is, the full chains of called (calling) entities are
   --  already computed for these nodes). In the latter case, the full chain
   --  of called (calling) entities is computed for the argument node and
   --  the corresponding flag (All_Calls_Set or All_Callers_Set) is set on.
   --  In the very end this procedure checks if both these flags
   --  (All_Calls_Set and All_Callers_Set) are ON, and if they are, sets
   --  All_CG_Details_Set also ON.

   function Check_Chain
     (GS_Node : GS_Node_Id;
      Callers : Boolean)
      return    Boolean;
   --  Checks if for the argument node the corresponding call chain (callers
   --  chain if Callers is ON, or called chain otherwise) consists only from
   --  the entities for that All_Callers_Set (All_Calls_Set) is ON.

   procedure Close_Recursive_Nodes;
   --  This is a part of Closure_Step, performed when all the "unclosed" call
   --  chains contains recursive calls. It detects the first "unclosed" and,
   --  therefore, having recursive calls node and sets the lists of all
   --  calls and all callers for it (setting the All_Calls_Set, All_Callers_Set
   --  and All_CG_Details_Set ON).

   procedure  Close_Recursive_Node (N :  GS_Node_Id; Callers : Boolean);
   --  Provided that N is a "recursive" node in a sense described in the
   --  documentation of Close_Recursive_Nodes routine, sets the lists of all
   --  calls or all callers for it, depending on Caller parameter

   procedure Close_List (N :  GS_Node_Id; Callers : Boolean);
   --  This procedure is called as a part of analyzing the node during
   --  the transitive closure of the call graph. If Callers is set ON, it
   --  sets the All_Callers_Chain list of the N by adding to this list all the
   --  nodes from All_Callers_Chain lists from the nodes that directly calls
   --  N (note, that for all these direct callers All_Callers_Set should
   --  already be ON. If Callers is OFF, the same happens for All_Calls_Chain
   --  list and called nodes.

   procedure Build_Recursive_Chain (For_Node : GS_Node_Id);
   --  Provided that N is a recursive node, creates in Recursive_Chain table a
   --  list  of nodes that recursively call each other. The list starts from N.
   --  This procedure is supposed to be called when we already have the lists
   --  of all calls/callers
   --  ??? Should this pocedure be moved into package interface?

   function Get_Worst_SE return GS_Node_Id;
   --  For the nodes stored in Recursive_Chain table (this function suppeses
   --  that the Recursive_Chain table contains at least one node stored)
   --  defines and returns as a result the node having the most serious side
   --  effect.

   ----------------------
   --  Recursive chain --
   ----------------------

   --  The following table is used when resolving full calls/callers lists for
   --  a recursive node. When during filling in the list of the all
   --  calls/callers of some node we detect a node that is not "closed" in the
   --  call graph closure sense, we put this node in the chain to avoid cycling

   package Recursive_Chain  is new Table.Table
     (Table_Component_Type => GS_Node_Id,
      Table_Index_Type     => Natural,
      Table_Low_Bound      => 1,
      Table_Initial        => 100,
      Table_Increment      => 100,
      Table_Name           => "recursive chain");

   function Already_In_Recursive_Chain (N : GS_Node_Id) return Boolean;
   --  Checks if N is already added to the recursive chain

   procedure Add_To_Recursive_Chain (N : GS_Node_Id);
   --  Adds N to the recursive chain

   -----------------
   -- Add_CG_Info --
   -----------------

   procedure Add_CG_Info
     (Element :        Asis.Element;
      Control : in out Traverse_Control;
      State   : in out Rule_Traversal_State)
   is
      Tmp : GS_Node_Id;
   begin

      --  Is_Scope (Element) can also have elaboration calls, so we have to
      --  process elaboration calls in a separate IF statement
      if Can_Have_Elab_Calls (Element) then
         Process_Elab_Calls (Element, Control, State);
      end if;

      if Is_Exec_Call (Element) then
         Process_Call (Element);
      elsif Is_Task_Creation (Element) then
         Store_Task_Creation_Arc (Element);
      elsif Is_Scope (Element) then
         Tmp := Register_Entity (Element);
         Set_Current_Scope (Tmp);
         GS_Nodes.Table (Tmp).Is_Scope := True;
      elsif Is_Caller (Element)
        or else
            Definition_Kind (Element) in
              A_Task_Definition .. A_Protected_Definition
      then
         Tmp := Register_Entity  (Element, Fix_Protected_Op => True);
         Set_Enclosing_Scope (Tmp);
      elsif Is_Renaming (Element) then
         Process_Renaming (Element);
      end if;

   end Add_CG_Info;

   ----------------------------
   -- Add_To_Recursive_Chain --
   ----------------------------

   procedure Add_To_Recursive_Chain (N : GS_Node_Id) is
   begin
      --  A rather simple approach, no checks at all.
      Recursive_Chain.Append (N);
   end Add_To_Recursive_Chain;

   --------------------------------
   -- Already_In_Recursive_Chain --
   --------------------------------

   function Already_In_Recursive_Chain (N : GS_Node_Id) return Boolean is
      Result : Boolean := False;
   begin
      --  A rather simple approach, we hope that recursive chains are
      --  relatively short, so we are not expecting any performance problems
      --  here:

      for J in Recursive_Chain.First .. Recursive_Chain.Last loop

         if Recursive_Chain.Table (J) = N then
            Result := True;
            exit;
         end if;

      end loop;

      return Result;
   end Already_In_Recursive_Chain;

   ------------------
   -- Analyze_Node --
   ------------------

   procedure Analyze_Node (N : GS_Node_Id; Status_Changed : out Boolean) is
      Current_Node : GS_Node_Record renames GS_Nodes.Table (N);
      All_Closed   : Boolean;
   begin
      Status_Changed := False;

      if not Current_Node.All_Callers_Set then

         if No (Current_Node.Callers_Chain) then
            Current_Node.All_Callers_Set := True;
            Status_Changed               := True;
            --  Nothing else to do!
         else
            All_Closed := Check_Chain (N, Callers => True);

            if All_Closed then
               Close_List (N, Callers => True);
               Current_Node.All_Callers_Set := True;
               Status_Changed               := True;
            end if;

         end if;

      end if;

      if not Current_Node.All_Calls_Set then

         if No (Current_Node.Calls_Chain) then
            Current_Node.All_Calls_Set := True;
            Status_Changed             := True;
            --  Nothing else to do!
         else
            All_Closed := Check_Chain (N, Callers => False);

            if All_Closed then
               Close_List (N, Callers => False);
               Current_Node.All_Calls_Set := True;
               Status_Changed             := True;
            end if;

         end if;

      end if;

      Current_Node.All_CG_Details_Set :=
        Current_Node.All_Callers_Set and then Current_Node.All_Calls_Set;

   end Analyze_Node;

   ---------------------------
   -- Build_Recursive_Chain --
   ---------------------------

   procedure Build_Recursive_Chain (For_Node : GS_Node_Id) is
      Next_Call        : Link_Id    := First_Called_Entity (For_Node);
      Next_Called_Node : GS_Node_Id := Node (Next_Call);
   begin
      Recursive_Chain.Init;
      Recursive_Chain.Append (For_Node);

      while Present (Next_Called_Node) loop

         if Next_Called_Node /= For_Node
           and then
            In_List (For_Node, First_Called_Entity (Next_Called_Node))
         then
            Recursive_Chain.Append (Next_Called_Node);
         end if;

         Next_Call        := Next (Next_Call);
         Next_Called_Node := Node (Next_Call);
      end loop;

   end Build_Recursive_Chain;

   ---------------------------
   -- Check_CG_Completeness --
   ---------------------------

   procedure Check_CG_Completeness is
   begin

      --  First, set Body_Analyzed for task objects abd protected operations
      --  in case if the corresponding body is processed (for tasks and
      --  protected objects declared by an object declarations)

      for J in First_GS_Node .. GS_Nodes.Last loop

         if (GS_Node_Kind (J) = A_Task_Object
            and then
             Body_Analyzed (Get_Task_Definition (J)))
           or else
            (GS_Node_Kind (J) in A_Protected_Procedure .. A_Protected_Entry
            and then
             Body_Analyzed (Get_Protected_Op (J)))
         then
            Set_Body_Analyzed (J, True);
         end if;

      end loop;

      for J in First_GS_Node .. GS_Nodes.Last loop

         if not Body_Analyzed (J) then
            Error ("body is not analyzed for " & Get_String (Location (J)));
         end if;

      end loop;

   end Check_CG_Completeness;

   -----------------
   -- Check_Chain --
   -----------------

   function Check_Chain
     (GS_Node : GS_Node_Id;
      Callers : Boolean)
      return    Boolean
   is
      Next_Link : Link_Id;
      Next_N    : GS_Node_Id := No_GS_Node;
      Result    : Boolean := True;

      function Node_Closed (N : GS_Node_Id) return Boolean;
      --  Checks if the callers or calls (depending on the value of Caller
      --  parameter) chain for the argument node is fully defined

      function Node_Closed (N : GS_Node_Id) return Boolean is
      begin
         if Callers then
            return GS_Nodes.Table (N).All_Callers_Set;
         else
            return GS_Nodes.Table (N).All_Calls_Set;
         end if;
      end Node_Closed;

   begin

      if Callers then
         Next_Link := First_Directly_Calling_Entity (GS_Node);
      else
         Next_Link := First_Directly_Called_Entity (GS_Node);
      end if;

      while Present (Next_Link) loop

         Next_N := Node (Next_Link);

         if not Node_Closed (Next_N) then
            Result := False;
            exit;
         end if;

         Next_Link := Next (Next_Link);
      end loop;

      return Result;
   end Check_Chain;

   ----------------
   -- Close_List --
   ----------------

   procedure Close_List (N :  GS_Node_Id; Callers : Boolean) is
      Next_Link : Link_Id;
      Next_Node : GS_Node_Id;
   begin
      --  First, copy all the direct callers/called entities, no check is
      --  needed here.

      if Callers then
         Copy_List
           (To   => GS_Nodes.Table (N).All_Callers_Chain,
            From => GS_Nodes.Table (N).Callers_Chain);
      else
         Copy_List
           (To   => GS_Nodes.Table (N).All_Calls_Chain,
            From => GS_Nodes.Table (N).Calls_Chain);
      end if;

      --  Now, add entities from the all callers/called entities lists of all
      --  the direct callers/called entities.

      --  Note, that the current implementation has very bad performance!!!

      if Callers then
         Next_Link := First_Directly_Calling_Entity (N);
      else
         Next_Link := First_Directly_Called_Entity (N);
      end if;

      while Present (Next_Link) loop
         Next_Node := Node (Next_Link);

         if Callers then
            Add_From_List
              (To   => GS_Nodes.Table (N).All_Callers_Chain,
               From => GS_Nodes.Table (Next_Node).All_Callers_Chain);
         else
            Add_From_List
              (To   => GS_Nodes.Table (N).All_Calls_Chain,
               From => GS_Nodes.Table (Next_Node).All_Calls_Chain);
         end if;

         Next_Link := Next (Next_Link);
      end loop;
   end Close_List;

   --------------------------
   -- Close_Recursive_Node --
   --------------------------

   procedure  Close_Recursive_Node (N :  GS_Node_Id; Callers : Boolean) is
      Next_Arc : Link_Id;

      function First_Arc
        (N       : GS_Node_Id;
         Callers : Boolean)
         return    Link_Id;
      --  Returns the beginning (the first arc) of the list of direct calls
      --  or callers, depending on the value of Callers

      procedure Add_Nodes
        (To      : GS_Node_Id;
         From    : GS_Node_Id;
         Callers : Boolean);
      --  Depending on the Callers, add to the list of all calls (callers) of
      --  To the From node and all the calls, generated from the node From. The
      --  point here is that for the node From the list of all calls may not be
      --  set yet, so this routine goes through the list of direct calls of
      --  From, and for closed nodes from this list just adds lists of their
      --  full calls, and processes the unclosed (recursive) nodes recursively.
      --  To avoid circling, such "unclosed" node is put in the recursive chain
      --  table and is not processed when we visit it once again

      procedure Add_Nodes
        (To      : GS_Node_Id;
         From    : GS_Node_Id;
         Callers : Boolean)
      is

         List_To_Add_From : constant Link_Id := First_Arc (From, Callers);
         --  The direct calls or direct callers list of From node,
         --  depending on Callers

         Arc       : Link_Id;
         Next_Node : GS_Node_Id;
         --  Used to traverse List_To_Add_From

         Local_Arc       : Link_Id;
         Next_Local_Node : GS_Node_Id;
         --  Used to traverse the corresponding list of each node visited
         --  during traversing List_To_Add_From
      begin

         if (Callers
           and then
            GS_Nodes.Table (From).All_Callers_Set)
           or else
            (not Callers
           and then
            GS_Nodes.Table (From).All_Calls_Set)
         then

            --  First - is the node already closed? If it is - just add it and
            --  all it calls/callers

            if Callers then
               Add_To_List
                 (To   => GS_Nodes.Table (N).All_Callers_Chain,
                  N    => From);
            else
               Add_To_List
                 (To   => GS_Nodes.Table (N).All_Calls_Chain,
                  N    => From);
            end if;

            if Callers then
               Add_From_List
                 (To   => GS_Nodes.Table (N).All_Callers_Chain,
                  From => GS_Nodes.Table (From).All_Callers_Chain);
            else
               Add_From_List
                 (To   => GS_Nodes.Table (N).All_Calls_Chain,
                  From => GS_Nodes.Table (From).All_Calls_Chain);
            end if;

         else
            --  This is a recursive node. It may have some calls/callers in
            --  the direct calls/callers lists completely set, and some other
            --  may also be recursive.

            if not Already_In_Recursive_Chain (From) then
               --  This check is to avoid cycling when traversing recursive
               --  calls

               Add_To_Recursive_Chain (From);
               --  We will not process this node again after that

               --  First, add the node itself
               if Callers then
                  Add_To_List
                    (To   => GS_Nodes.Table (N).All_Callers_Chain,
                     N    => From);
               else
                  Add_To_List
                    (To   => GS_Nodes.Table (N).All_Calls_Chain,
                     N    => From);
               end if;

               --  And now - traversing the list of direct calls/callers:

               Arc := List_To_Add_From;

               while Present (Arc) loop
                  Next_Node := Node (Arc);

                  if (Callers
                    and then
                     GS_Nodes.Table (Next_Node).All_Callers_Set)
                    or else
                     (not Callers
                    and then
                     GS_Nodes.Table (Next_Node).All_Calls_Set)
                  then
                     --  For the given node the full call information is known,
                     --  we can just add the node and the list of all its
                     --  calls/callers to the result, there is no recursion
                     --  here

                     if Callers then
                        Add_To_List
                          (To   => GS_Nodes.Table (N).All_Callers_Chain,
                           N    => Next_Node);
                     else
                        Add_To_List
                          (To   => GS_Nodes.Table (N).All_Calls_Chain,
                           N    => Next_Node);
                     end if;

                     if Callers then
                        Add_From_List
                          (To   => GS_Nodes.Table (N).All_Callers_Chain,
                           From =>
                             GS_Nodes.Table (Next_Node).All_Callers_Chain);
                     else
                        Add_From_List
                          (To   => GS_Nodes.Table (N).All_Calls_Chain,
                           From => GS_Nodes.Table (Next_Node).All_Calls_Chain);
                     end if;

                  else
                     --  Next_Node is in a recursive call chain

                     --  Note, that in case of empty calls/callers lists
                     --  All_Calls_Set/All_Callers_Set flag should be OFF here!
                     --  ???

                     if not Already_In_Recursive_Chain (Next_Node) then

                        Add_To_Recursive_Chain (Next_Node);
                        --  To avoid infinite recursing!

                        --  We now for sure that Next_Node cannot be in
                        --  List_To_Close, so we can use something more
                        --  effective here!!!???

                        if Callers then
                           Add_To_List
                             (To   => GS_Nodes.Table (N).All_Callers_Chain,
                              N    => Next_Node);
                        else
                           Add_To_List
                             (To   => GS_Nodes.Table (N).All_Calls_Chain,
                              N    => Next_Node);
                        end if;

                        --  And now, we have to go to the chain of direct
                        --  calls/callers of this list and add nodes from it:
                        Local_Arc := First_Arc (Next_Node, Callers);
                        pragma Assert (Present (Local_Arc));

                        while Present  (Local_Arc) loop
                           Next_Local_Node := Node (Local_Arc);

                           Add_Nodes
                             (To      => To,
                              From    => Next_Local_Node,
                              Callers => Callers);

                           Local_Arc := Next (Local_Arc);
                        end loop;

                     end if;

                  end if;

                  Arc := Next (Arc);
               end loop;

            end if;

         end if;

      end Add_Nodes;

      function First_Arc
        (N       : GS_Node_Id;
         Callers : Boolean)
         return    Link_Id
      is
         Result : Link_Id;
      begin
         if Callers then
            Result := First_Directly_Calling_Entity (N);
         else
            Result := First_Directly_Called_Entity (N);
         end if;

         return Result;
      end First_Arc;

   begin -- Close_Recursive_Node

      Recursive_Chain.Init;

      Next_Arc := First_Arc (N, Callers);
      pragma Assert (Present (Next_Arc));
      --  We are iterating through the DIRECT calls/callers of N!

      while Present (Next_Arc) loop
         Add_Nodes (To => N, From => Node (Next_Arc), Callers => Callers);
         Next_Arc := Next (Next_Arc);
      end loop;

      if Callers then
         GS_Nodes.Table (N).All_Callers_Set := True;
      else
         GS_Nodes.Table (N).All_Calls_Set := True;
      end if;

   end Close_Recursive_Node;

   ---------------------------
   -- Close_Recursive_Nodes --
   ---------------------------

   procedure Close_Recursive_Nodes is
      Recursive_Node : GS_Node_Id;
      --  The recursive node to close

      function Get_Recursive_Node return GS_Node_Id;
      --  Returns the first non-closed recursive node for the current state of
      --  the node table. This function is supposed to be called when all the
      --  non-recursive nodes have already processed, so only for non-processed
      --  recursive node at least one of the fields All_Callers_Set or
      --  All_Callers_Set can be true

      function Get_Recursive_Node return GS_Node_Id is
         Result : GS_Node_Id := No_GS_Node;
      begin
         --  This function is extremely uneffective!
         for J in First_GS_Node .. GS_Nodes.Last loop

            if not GS_Nodes.Table (J).All_CG_Details_Set then
               Result := J;
               exit;
            end if;

         end loop;

         return Result;
      end Get_Recursive_Node;

   begin
      Recursive_Node := Get_Recursive_Node;
      pragma Assert (Present (Recursive_Node));

      if not GS_Nodes.Table (Recursive_Node).All_Callers_Set then
         Close_Recursive_Node (Recursive_Node, Callers => True);
      end if;

      if not GS_Nodes.Table (Recursive_Node).All_Calls_Set then
         Close_Recursive_Node (Recursive_Node, Callers => False);
      end if;

      GS_Nodes.Table (Recursive_Node).All_CG_Details_Set := True;
   end Close_Recursive_Nodes;

   ------------------
   -- Closure_Step --
   ------------------

   procedure Closure_Step (Remains_Unclosed_Node : out Boolean) is
      No_Nonrecursive_Closure : Boolean := True;
      --  This flag is set OFF if during the iteration through the nodes stored
      --  in the node table we were able to change at least something in the
      --  closure state of at least one node. If this flag remains ON after
      --  finishing the iteration, this means that all that we still have
      --  "unclousured" in the table are groups of mutually recursive entities.

      Status_Changed : Boolean;
   begin
      Remains_Unclosed_Node := False;

      --  First, try to make any local closure steps for nonrecursive nodes.

      for J in First_GS_Node .. GS_Nodes.Last loop

         if not GS_Nodes.Table (J).All_CG_Details_Set then
            Analyze_Node (J, Status_Changed);

            if Status_Changed then
               No_Nonrecursive_Closure := False;
            end if;

            if not GS_Nodes.Table (J).All_CG_Details_Set then
               Remains_Unclosed_Node := True;
            end if;

         end if;

      end loop;

      if Remains_Unclosed_Node and then No_Nonrecursive_Closure then
         Close_Recursive_Nodes;
      end if;

   end Closure_Step;

   ----------------------
   -- Complete_CG_Info --
   ----------------------

   procedure Complete_CG_Info (El : Asis.Element) is
   begin

      if Is_Scope (El) then
         Remove_Current_Scope;
      end if;

   end Complete_CG_Info;

   ------------------
   -- Get_Worst_SE --
   ------------------
   function Get_Worst_SE return GS_Node_Id is
      Result           : GS_Node_Id           :=
         Recursive_Chain.Table (Recursive_Chain.First);
      Current_SE       : Side_Effect_Statuses := Side_Effect_Status (Result);
      Current_SE_Level : Scope_Levels;
   begin
      if Current_SE = Local_Side_Effect then
         Current_SE_Level := Local_Side_Effect_Level (Result);
      end if;

      for J in Recursive_Chain.First + 1 .. Recursive_Chain.Last loop

         if Current_SE < Side_Effect_Status (Recursive_Chain.Table (J))
           or else
            (Current_SE = Local_Side_Effect
            and then
             Current_SE = Side_Effect_Status (Recursive_Chain.Table (J))
            and then
             Current_SE_Level >
             Local_Side_Effect_Level (Recursive_Chain.Table (J)))
         then
            Result     := Recursive_Chain.Table (J);
            Current_SE := Side_Effect_Status (Result);

            if Current_SE = Local_Side_Effect then
               Current_SE_Level := Local_Side_Effect_Level (Result);
            end if;

         end if;
      end loop;

      return Result;
   end Get_Worst_SE;

   ---------------
   -- Is_Called --
   ---------------

   function Is_Called (N : GS_Node_Id) return Boolean is
   begin
      pragma Assert (Present (N));

      return GS_Nodes.Table (N).Is_Used;

--      return Present (First_Directly_Calling_Entity (N));
   end Is_Called;

   -----------------------
   -- Is_Recursive_Node --
   -----------------------

   function Is_Recursive_Node (N : GS_Node_Id) return Boolean is
   begin
      pragma Assert (Present (N));
      pragma Assert (GS_Nodes.Table (N).All_Calls_Set);

      return In_List (N, First_Called_Entity (N));
   end Is_Recursive_Node;

   ------------------
   -- Process_Call --
   ------------------

   procedure Process_Call (Call : Asis.Element) is
   begin

      if Is_Task_Entry_Call (Call) then
         Process_Task_Entry_Call (Call);
      elsif Is_Protected_Operation_Call (Call) then
         Process_Protected_Operation_Call (Call);
      else
         Process_Ordinary_Call (Call);
      end if;
   end Process_Call;

   ------------------------
   -- Process_Elab_Calls --
   ------------------------

   procedure Process_Elab_Calls
     (El      : Asis.Element;
      Control : in out Traverse_Control;
      State   : in out Rule_Traversal_State)
   is
      Type_Decl_El : Asis.Element := Nil_Element;
      --  To be set to point to the (full) type declaration of the type
      --  for that we have to process default (sub)component initialization
      --  expressions

      Process_Discriminants : Boolean := False;
      --  In case if the discriminant constraint is present, we do not have to
      --  process default expressions for discriminants

      procedure Process_Type_Elab_Code
        (El                    : Asis.Element;
         Process_Discriminants : Boolean);
      --  Collects the call graph information from all the possible component
      --  initialization expression, recursevely traversing the type structure
      --  of the type declaration represented by El. Initialization expressions
      --  for discrimitatnts are processed only if Process_Discriminants is ON.

      procedure Process_Type_Elab_Code
        (El                    : Asis.Element;
         Process_Discriminants : Boolean)
      is
         Type_Components : constant Asis.Element_List :=
           Get_Type_Components (El, Process_Discriminants);
         --  List of the component declarations elaborated as a part of the
         --  elaborating of the type declaration

         Expr : Asis.Element;
         --  Initialization expression of the componment

         Comp_Type : Asis.Element;
         --  If we have an array, we have to investigate the component type as
         --  well
      begin

         if Is_Nil (Type_Components) then

            --  The only possibility that can be of interest for us is an array
            --  type with record components (including possible derivations)

            Comp_Type := Type_Declaration_View (El);

            if Type_Kind (Comp_Type) = A_Derived_Type_Definition then
               --  Note, that we do not care about
               --  A_Derived_Record_Extension_Definition here, because
               --  for them we should get both explicit and implicit components
               --  from the call to Get_Type_Components. We also do not care
               --  about possible discriminant constraints here

               Comp_Type := Corresponding_Root_Type (Comp_Type);
               Comp_Type :=  Type_Declaration_View (Comp_Type);
            end if;

            if Type_Kind (Comp_Type) in
               An_Unconstrained_Array_Definition ..
               A_Constrained_Array_Definition
            then
               Comp_Type :=
                 Component_Subtype_Indication
                   (Array_Component_Definition (Comp_Type));

               Comp_Type := Asis.Definitions.Subtype_Mark (Comp_Type);

               if Expression_Kind (Comp_Type) /= An_Attribute_Reference then
                  Comp_Type := Get_Type_Decl_From_Subtype_Mark (Comp_Type);

                  if Declaration_Kind (Comp_Type) /=
                     A_Task_Type_Declaration
                  then
                     --  It can be no elaboration code for task type!
                     Process_Type_Elab_Code
                       (Comp_Type, Process_Discriminants => False);

                  end if;

               end if;

            end if;

         else

            for J in Type_Components'Range loop
               Expr := Initialization_Expression (Type_Components (J));

               if Is_Nil (Expr) then
                  --  But the component type may contain initialization
                  --  expressions! The component itself should have a definite
                  --  subtype, so there is no need to analyse component
                  --  discriminants
                  Comp_Type :=
                    Component_Subtype_Indication
                      (Object_Declaration_View (Type_Components (J)));

                  Comp_Type := Asis.Definitions.Subtype_Mark (Comp_Type);

                  if Expression_Kind (Comp_Type) /= An_Attribute_Reference then
                     Comp_Type := Get_Type_Decl_From_Subtype_Mark (Comp_Type);

                     if Declaration_Kind (Comp_Type) /=
                        A_Task_Type_Declaration
                     then
                        --  It can be no elaboration code for task type!
                        Process_Type_Elab_Code
                          (Comp_Type, Process_Discriminants => False);

                     end if;

                  end if;

               else
                  Gnatcheck.Rules.Traversing.Extract_Global_Information
                    (Element => Expr,
                     Control => Control,
                     State   => State);
               end if;

            end loop;

         end if;

      end Process_Type_Elab_Code;

   begin

      case Flat_Element_Kind (El) is
         when A_Variable_Declaration =>

            if Is_Nil (Initialization_Expression (El)) then

               Type_Decl_El := Object_Declaration_View (El);

               if Definition_Kind (Type_Decl_El) = A_Type_Definition  then
                  --  The case of A : array (...) of ...
                  Type_Decl_El :=
                    Component_Subtype_Indication
                      (Array_Component_Definition (Type_Decl_El));

                  --  Components should have a definite subtype, so
                  Process_Discriminants := False;
               else
                  Process_Discriminants :=
                     Is_Nil (Subtype_Constraint (Type_Decl_El))
                    and then
                     Is_Indefinite_Subtype
                       (Asis.Definitions.Subtype_Mark (Type_Decl_El));
               end if;

               --  We can not have A_Component_Definition, A_Task_Definition or
               --  A_Protected_Definition case here, because we started from
               --  A_Variable_Declaration so now A_Subtype_Indication is the
               --  only possibility for Type_Decl_El

               Type_Decl_El := Asis.Definitions.Subtype_Mark (Type_Decl_El);

               if Expression_Kind (Type_Decl_El) /=
                     An_Attribute_Reference
               then
                  --  In case of a attribute reference as a subtype mark the
                  --  only possible case is 'Base, so we have a scalar type
                  --  here, therefore it can be no default initialization

                  Type_Decl_El :=
                    Get_Type_Decl_From_Subtype_Mark (Type_Decl_El);

                  if Declaration_Kind (Type_Decl_El) /=
                     A_Task_Type_Declaration
                  then
                     --  It can be no elaboration code for task type!
                     Process_Type_Elab_Code
                       (Type_Decl_El, Process_Discriminants);

                  end if;

               end if;

            end if;

--         when  =>
         when others =>
            --  For all the other cases - Placeholder for the moment !!! ???
            null;
      end case;

   end Process_Elab_Calls;

   ---------------------------
   -- Process_Ordinary_Call --
   ---------------------------

   procedure Process_Ordinary_Call (Call : Asis.Element) is
      Called_El : Asis.Element := Get_Called_Element (Call);
   begin
      if Is_Nil (Called_El) then

         if Is_Call_To_Predefined_Operation (Call)
           or else
            Is_Call_To_Attribute_Subprogram (Call)
         then
            --  We do not consider such calls at all
            return;
         elsif Is_Dispatching_Call (Call) then
            Store_Dispatching_Call;
         elsif Is_Dynamic_Call (Call) then
            Store_Dynamic_Call;
         else
            pragma Assert (False);
            null;
         end if;

      elsif Declaration_Kind (Called_El) =
            An_Enumeration_Literal_Specification
      then
         --  This may happen in instantiation if an enumeration literal is
         --  used as an actual for a formal function.
         return;
      else
         Called_El := Corresponding_Element (Called_El);

         if not Is_Predefined_Operation_Renaming (Called_El) then
            Store_Arc (Called_El);
         end if;

      end if;

   end Process_Ordinary_Call;

   --------------------------------------
   -- Process_Protected_Operation_Call --
   --------------------------------------

   procedure Process_Protected_Operation_Call (Call : Asis.Element) is
      Protectes_Op  : Asis.Element;
      Protectes_Obj : Asis.Element;
      --  In case of a protected operation call, we have to detect both the
      --  called protected operation and the protected object this operation
      --  belongs to.

      Protected_Op_Node : GS_Node_Id;
      --  Should be set to the body entity of the corresponding protected
      --  operation

      Called_Entity : GS_Node_Id;
      --  Should be set to the entity representing the concrete (that is -
      --  belonging to a specific protected object) protected operation that
      --  is called

   begin

      if Expression_Kind (Call) = A_Function_Call then
         Protectes_Op := Corresponding_Called_Function (Call);
      else
         Protectes_Op := Corresponding_Called_Entity (Call);
      end if;

      Protected_Op_Node := Corresponding_Node (Protectes_Op);

      Protectes_Obj := Corresponding_Protected_Object (Call);

      pragma Assert (not (Is_Nil (Protectes_Obj)));

      --  If Is_Nil (Protectes_Obj), then we have to store a dynamic (???) call
      --  here

      Called_Entity := Corresponding_Protected_Op_Node
        (Protectes_Obj, Protected_Op_Node);

      Store_Arc (Called_Entity);
   end Process_Protected_Operation_Call;

   ----------------------
   -- Process_Renaming --
   ----------------------

   procedure Process_Renaming (El : Asis.Element) is
      Renaming_Node : GS_Node_Id;
      --  Node corresponding to this renaming

      Renamed_El : Asis.Element;
      --  In some cases we have to correct the renamed element.

      Renamed_Node : GS_Node_Id := No_GS_Node;
      --  Node corresponding to the renamed entity

      Renaming_Kind : Renaming_Kinds := Not_A_Renamimg;
   begin

      --  Partial implementation only!!!

      if Is_Predefined_Operation_Renaming (El) then
         return;
      end if;

      Renaming_Node := Corresponding_Node (Corresponding_Declaration (El));

      if Is_Enum_Literal_Renaming (El) then
         Renaming_Kind := Enum_Literal_Renaming;
         Set_Body_Analyzed (Renaming_Node, True);
      else

         Renamed_El := Declaration_Of_Renamed_Entity (El);

         --  Correcting the renamed element, the correction depends on the
         --  current approach to representation this or that entities in the
         --  call graph

         case Declaration_Kind (Renamed_El) is
            when An_Entry_Declaration =>
               if Definition_Kind (Enclosing_Element (Renamed_El)) =
                  A_Task_Definition
               then
                  Renamed_El :=
                    Enclosing_Element (Enclosing_Element (Renamed_El));
               end if;

--            when  =>
                  --  ... to be continued?
            when others =>
               null;
         end case;

         Renamed_Node := Corresponding_Node (Renamed_El);
         --  What about the case when Declaration_Of_Renamed_Entity returns
         --  Nil_Element?
      end if;

      if Is_Renaming_As_Body (El) then
         Set_Body_Analyzed (Renaming_Node, True);
         Renaming_Kind := Renaming_As_Body;
      elsif Renaming_Kind = Not_A_Renamimg then
         Set_Body_Analyzed (Renaming_Node, True);
         Renaming_Kind := Renaming_As_Declaration;
      end if;

      Set_Renamed_Entity (Renaming_Node,        Renamed_Node);
      Set_GS_Node_Renaming_Kind (Renaming_Node, Renaming_Kind);
      Set_Enclosing_Scope (Renaming_Node);

   end Process_Renaming;

   -----------------------------
   -- Process_Task_Entry_Call --
   -----------------------------

   procedure Process_Task_Entry_Call (Call : Asis.Element) renames
     Process_Ordinary_Call;
   --  This is a temporary solution!!!???

   --------------------
   -- Set_Priorities --
   --------------------

   procedure Set_Priorities is
      procedure Set_Node_Priority (N : GS_Node_Id);
      --  Defines the priority for the argument node (it is supposed that this
      --  priority is not defined at the moment of the call). As a side effect
      --  can define priorities for the nodes upon that the priority of the
      --  given node depends upon

      procedure Set_Node_Priority (N : GS_Node_Id) is
         P : GS_Node_Id;
         --  This is the node to take (inherit) the priority from
      begin

         case GS_Node_Kind (N) is

            when A_Protected_Procedure .. A_Protected_Entry =>
               P := Get_Protected_Op (N);

            when A_Protected_Procedure_Body .. A_Protected_Entry_Body =>
               P := Enclosing_Protected_Definition (N);
            when A_Task_Object =>
               P := Get_Task_Definition (N);

            when others =>
               P := Enclosing_Scope (N);
         end case;

         if not Priority_Defined (P) then
            Set_Node_Priority (P);
         end if;

         Set_Priority (N, Node_Priority (P));

      end Set_Node_Priority;

   begin

      for J in First_GS_Node .. GS_Nodes.Last loop

         if not Priority_Defined (J) then
            Set_Node_Priority (J);
         end if;

      end loop;

   end Set_Priorities;

   ---------------------
   -- Set_Side_Effect --
   ---------------------

   procedure Set_Side_Effect is

      procedure Set_SE_Step (State_Changed : out Boolean);
      --  This procedure traverses all the call graph nodes and tries to define
      --  completely a side effect for a node on the base of analysing its
      --  called nodes. State_Changed is set ON if at least for one node
      --  it changes the Side_Effect_Defined status from OFF to ON, otherwise
      --  it is set OFF.

      procedure Resolve_Recursion;
      --  This procedure trues to resolve the side effect settings for
      --  recursive call chains. (That is, if P calls P, we do not know what id
      --  side effect status of P, because P calls a subprogram with unknown
      --  side effect. This routine tries to resolve such situations)

      procedure Set_Task_Objects_and_Renamings;
      --  This procedure copies into a task object the side effect data from
      --  the corresponding task definition, and in the renaming objects -
      --  the corresponding data from the renamed object

      State_Changed : Boolean := True;

      -----------------------
      -- Resolve_Recursion --
      -----------------------

      procedure Resolve_Recursion is
         procedure Resolve_SE_For_Node (N : GS_Node_Id);
         --  Resolves the side effect for the recursive chain containing N

         procedure Resolve_SE_For_Node (N : GS_Node_Id) is
            Worst_SE_Node   : GS_Node_Id;
         begin
            pragma Assert (Is_Recursive_Node (N));

            Build_Recursive_Chain (N);

            Worst_SE_Node := Get_Worst_SE;

            for J in 1 .. Recursive_Chain.Last loop
               Correct_Side_Effect_Status
                 (For_Node  => Recursive_Chain.Table (J),
                  From_Node => Worst_SE_Node);

               Set_Side_Effect_Defined (Recursive_Chain.Table (J));
            end loop;

            null;
         end Resolve_SE_For_Node;

      begin

         for J in First_GS_Node .. GS_Nodes.Last loop

            if not Side_Effect_Defined (J) then
               pragma Assert (Call_To_Unknown_SE (J));
               Resolve_SE_For_Node (J);
            end if;

         end loop;

      end Resolve_Recursion;

      -----------------
      -- Set_SE_Step --
      -----------------

      procedure Set_SE_Step (State_Changed : out Boolean) is
         Next_Caller            : Link_Id;
         State_Changed_For_Node : Boolean;
      begin

         State_Changed := False;

         for J in First_GS_Node .. GS_Nodes.Last loop

            if not Side_Effect_Defined (J) then

               case GS_Node_Renaming_Kind (J) is
                  when Enum_Literal_Renaming =>
                     --  It can be no side effect in this case!
                     Set_Side_Effect_Defined (J);
                     Set_Side_Effect_Status (J, No_Side_Effect);
                     State_Changed := True;

                  when Renaming_As_Body        |
                       Renaming_As_Declaration =>

                     if Side_Effect_Defined (Renamed_Entity (J)) then
                        Correct_Side_Effect_Status
                          (For_Node  => J,
                           From_Node => Renamed_Entity (J));

                        Set_Side_Effect_Defined (J);
                        State_Changed := True;
                     end if;

                  when Not_A_Renamimg =>

                     if GS_Node_Kind (J) = A_Task_Object then

                        if Side_Effect_Defined (Get_Task_Definition (J)) then
                           Correct_Side_Effect_Status
                             (For_Node  => J,
                              From_Node => Get_Task_Definition (J));

                           Set_Side_Effect_Defined (J);
                           State_Changed := True;
                        end if;

                     else

                        pragma Assert (Call_To_Unknown_SE (J));

                        State_Changed_For_Node := True;

                        Next_Caller := First_Called_Entity (J);

                        while Present (Next_Caller) loop

                           if Side_Effect_Defined (Node (Next_Caller)) then
                              Correct_Side_Effect_Status
                                (For_Node  => J,
                                 From_Node => Node (Next_Caller));
                           elsif J /= Node (Next_Caller) then
                              State_Changed_For_Node := False;
                              exit;
                           end if;

                           Next_Caller := Next (Next_Caller);
                        end loop;

                        if State_Changed_For_Node then
                           State_Changed := True;
                           Set_Side_Effect_Defined (J);
                           Set_Call_To_Unknown_SE (J, False);
                        end if;

                     end if;

               end case;

            end if;

         end loop;

      end Set_SE_Step;

      ------------------------------------
      -- Set_Task_Objects_and_Renamings --
      ------------------------------------

      procedure Set_Task_Objects_and_Renamings is
         procedure Copy_Renamed_SE (N : GS_Node_Id);
         --  Copies the side effect through the renaming chain ended with N

         procedure Copy_Renamed_SE (N : GS_Node_Id) is
            R : constant GS_Node_Id := Renamed_Entity (N);
         begin

            if GS_Node_Renaming_Kind (R) /= Not_A_Renamimg then
               Copy_Renamed_SE (R);
            end if;

            Correct_Side_Effect_Status
              (For_Node  => N,
               From_Node => R);

            Set_Side_Effect_Defined (N, Side_Effect_Defined (R));

         end Copy_Renamed_SE;

      begin
         for J in First_GS_Node .. GS_Nodes.Last loop

            if not Side_Effect_Defined (J) then

               if GS_Node_Kind (J) = A_Task_Object then
                  Correct_Side_Effect_Status
                    (For_Node  => J,
                     From_Node => Get_Task_Definition (J));

                  Set_Side_Effect_Defined
                   (J, Side_Effect_Defined (Get_Task_Definition (J)));

               elsif GS_Node_Renaming_Kind (J) /= Not_A_Renamimg then
                  Copy_Renamed_SE (J);
               end if;

            end if;

         end loop;

      end Set_Task_Objects_and_Renamings;

   begin

      while State_Changed loop
         Set_SE_Step (State_Changed);
      end loop;

      --  Set_Task_Objects_and_Renamings;

      Resolve_Recursion;

      Set_Task_Objects_and_Renamings;

      for J in First_GS_Node .. GS_Nodes.Last loop
         --  For the development period only!
         pragma Assert (Side_Effect_Defined (J));
         null;
      end loop;

   end Set_Side_Effect;

   -----------------------
   -- Set_Used_Entities --
   -----------------------

   procedure Set_Used_Entities is

      procedure Set_Used_Step (State_Changed : out Boolean);
      --  This procedure traverses all the call graph nodes and sets for a node
      --  Is_Used flag ON. State_Changed is set ON if at least for one node
      --  Is_Used flag is set ON, otherwise it is set OFF.

      State_Changed : Boolean := True;

      procedure Set_Used_Step (State_Changed : out Boolean) is
         Next_Caller : Link_Id;
      begin

         State_Changed := False;

         for J in First_GS_Node .. GS_Nodes.Last loop

            if not GS_Nodes.Table (J).Is_Used then

               Next_Caller := First_Calling_Entity (J);

               while Present (Next_Caller) loop

                  if GS_Nodes.Table (Node (Next_Caller)).Is_Used then
                     GS_Nodes.Table (J).Is_Used := True;
                     State_Changed              := True;
                     exit;
                  end if;

                  Next_Caller := Next (Next_Caller);
               end loop;

            end if;

         end loop;

      end Set_Used_Step;

   begin

      while State_Changed loop
         Set_Used_Step (State_Changed);
      end loop;

   end Set_Used_Entities;

   ----------------------------
   -- Store_Dispatching_Call --
   ----------------------------

   procedure Store_Dispatching_Call is
   begin
      Set_Contains_Dispatching_Call (Current_Scope);
   end Store_Dispatching_Call;

   ------------------------
   -- Store_Dynamic_Call --
   ------------------------

   procedure Store_Dynamic_Call is
   begin
      Set_Contains_Dynamic_Call (Current_Scope);
   end Store_Dynamic_Call;

   ------------------------
   -- Transitive_Closure --
   ------------------------

   procedure Transitive_Closure is
      Remains_Unclosed_Node : Boolean := True;
   begin
      Check_CG_Completeness;

      Traverse_Renamings;

      while Remains_Unclosed_Node loop
         Closure_Step (Remains_Unclosed_Node);
      end loop;
   end Transitive_Closure;

   ------------------------
   -- Traverse_Renamings --
   ------------------------

   procedure Traverse_Renamings is
   begin
      --  !!! only prototype implementation !!! ???

      for J in First_GS_Node .. GS_Nodes.Last loop

         if GS_Node_Kind (J) in A_Procedure .. A_Function
           and then
            Present (Renamed_Entity (J))
         then
            --  renaming-as-body
            Copy_List
              (To   => GS_Nodes.Table (J).Calls_Chain,
               From => First_Called_Entity (Renamed_Entity (J)));

            Add_To_List
              (To => GS_Nodes.Table (J).Calls_Chain,
               N  => Renamed_Entity (J));

            Add_To_List
              (To => GS_Nodes.Table (Renamed_Entity (J)).Callers_Chain,
               N  => J);

         end if;

      end loop;

   end Traverse_Renamings;

end Gnatcheck.Global_State.CG;
