------------------------------------------------------------------------------
--                                                                          --
--                          GNATCHECK COMPONENTS                            --
--                                                                          --
--               G N A T C H E C K . G L O B A L _ S T A T E                --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                       Copyright (C) 2005, AdaCore                        --
--                                                                          --
-- GNATCHECK  is  free  software;  you can redistribute it and/or modify it --
-- under terms of the  GNU  General Public License as published by the Free --
-- Software Foundation;  either version 2, or ( at your option)  any  later --
-- version.  GNATCHECK  is  distributed in the hope that it will be useful, --
-- but  WITHOUT  ANY  WARRANTY;   without  even  the  implied  warranty  of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details.  You should have received a copy of the --
-- GNU  General Public License distributed with GNAT;  see file COPYING. If --
-- not,  write to the  Free Software Foundation,  51 Franklin Street, Fifth --
-- Floor, Boston, MA 02110-1301, USA.                                       --
--                                                                          --
-- GNATCHECK is maintained by AdaCore (http://www.adacore.com).             --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Characters.Handling;    use Ada.Characters.Handling;
with Ada.Text_IO;                use Ada.Text_IO;
with Ada.Wide_Text_IO;

with GNAT.OS_Lib;                use GNAT.OS_Lib;

with Asis.Compilation_Units;     use Asis.Compilation_Units;
with Asis.Declarations;          use Asis.Declarations;
with Asis.Elements;              use Asis.Elements;
with Asis.Expressions;           use Asis.Expressions;
with Asis.Extensions;            use Asis.Extensions;
with Asis.Extensions.Flat_Kinds; use Asis.Extensions.Flat_Kinds;
with Asis.Statements;            use Asis.Statements;
with Asis.Text;                  use Asis.Text;

with ASIS_UL.Output;             use ASIS_UL.Output;

with Gnatcheck.ASIS_Utilities;   use Gnatcheck.ASIS_Utilities;
with Gnatcheck.Diagnoses;
with Gnatcheck.Options;          use Gnatcheck.Options;

package body Gnatcheck.Global_State is

   Ident_String : constant String := "   ";
   --  Used in the debug output of the global data structures

   --------------------------------
   -- Hash table for Nodes Table --
   --------------------------------

   Hash_Num : constant Integer := 2**16;  --  ???
   --  Number of headers in the hash table. There is no special reason in this
   --  choice.

   Hash_Max : constant Integer := Hash_Num - 1;
   --  Indexes in the hash header table run from 0 to Hash_Num - 1

   subtype Hash_Index_Type is Integer range 0 .. Hash_Max;
   --  Range of hash index values

   Hash_Table : array (Hash_Index_Type) of GS_Node_Id;
   --  The hash table is used to locate existing entries in the nodes table.
   --  The entries point to the first nodes table entry whose hash value
   --  matches the hash code. Then subsequent nodes table entries with the
   --  same hash code value are linked through the Hash_Link fields.

   function Hash (El : Asis.Element) return Hash_Index_Type; --  ???
   --  Compute hash code for its argument. At the moment we are using rather
   --  primitive hash function, this should be revised at some point

   -----------------------
   -- Local Subprograms --
   -----------------------

   function Is_Equal
     (GS_Node      : GS_Node_Id;
      El           : Asis.Element;
      Protected_Op : GS_Node_Id := No_GS_Node)
      return         Boolean;
   --  Checks that GS_Node points to the call graph node corresponding to the
   --  given Element. Returns False if No (GS_Node). For the Protected_Op
   --  parameter, see the explanation in the documentation of Register_Entity.

   function Define_GS_Node_Kind (El : Asis.Element) return GS_Node_Kinds;
   --  Defines which node kind corresponds to the given Element

   function Register_Node
     (El         : Asis.Element;
      Node_Kind  : GS_Node_Kinds := Not_A_Node)
      return GS_Node_Id;
   --  Registers the node corresponding to the argument Element and returns its
   --  Id as a result. The caller is responsible for making sure that the
   --  corresponding Element has not been registered yet. Moreover, this
   --  function should be called only for the nodes that are used as the base
   --  for representing entities in the Global Structure, that is, to the
   --  results of Gnatcheck.ASIS_Utilities.Corresponding_Element.
   --  If set to non-null value, Encl_Scope is used to initialize the
   --  corresponding field of the allocated node
   --  If set, Node_Kind is used to set the kind of the new node, otherwise
   --  the kind is detected from the Element argument.
   --  This function can never return No_CG_Node.

   procedure Print_Node_Lists_Table;
   --  Prints out low-level node list info

   procedure Store_Needed_Source (El : Asis.Element);
   --  Tries to detect and to store in the Source Table the (full normalized)
   --  name of the source file that is needed to get the full information about
   --  the given element El. If the source cannot be located and stored,
   --  generates the corresponding error message.

   ------------------------------------------
   -- Representing Tasks in the Call Graph --
   ------------------------------------------

   --  We consider each task object as a callable entity that is called in
   --  the place where the task is created, either statically or dynamically.

   --  For a task entity, we consider that the corresponding task body in the
   --  same way as a subprogram body (to detect entities called by this task
   --  entity). This body is considered as being boind to the corresponding
   --  task definition (but not to task object or task type declaration - to
   --  avoid the difference between single tasks and task types)

   --  Any entry call we consider as a call to the whole task object. This is
   --  a rather big simplification, probably too rough for the tasks our
   --  Call Graph is supposed to be used for. Probably we should consider for
   --  each task entry a set of calls that may be generated by the
   --  corresponding accept statements for this call.

   --  At the implementation level, for task definitions, we store a
   --  (non-callable!) entity that represents the task (with all the calls
   --  issued from the task body) and for every task object we store the
   --  reference to this "task definition entity" (including the task objects
   --  declared by single task declarations)

   --  ??? Dynamically created tasks?
   --  ??? Tasks that are components of other data structures?
   --  ??? Tasks as parameters

   function Corresponding_Task_Node
     (El            : Asis.Element;
      Single_Task   : Boolean    := True;
      Task_Def_Node : GS_Node_Id := No_GS_Node)
      return          GS_Node_Id;
   --  Version of the Corresponding_Node function from above that takes into
   --  account the specifics of representing tasks. Two additional parameters
   --  are used if we have to allocate a new task node. The boolean Single_Task
   --  flag is used to make the difference between a task declared by a single
   --  task declaration (in this case it has its "own" definition and body) and
   --  a task declared by an object declaration or created by allocators (when
   --  the task type is used to define the task object). If Single_Task is set
   --  OFF, Task_Def_Node should be set to the corresponding task definition
   --  node, otherwise Task_Def_Node is ignored.

   procedure Set_Protected_Op (For_Obj : GS_Node_Id; Pr_Op : GS_Node_Id);
   --  Sets for the node representing a protected operation belonging to a
   --  specific protected object the reference to the node representing the
   --  executable body of the corresponding protected operation.

   procedure Set_Enclosing_Protected_Definition
     (For_PO : GS_Node_Id;
      P_Def  : GS_Node_Id);
   --  Sets the reference to enclosing protected definition for a node
   --  representing the protected operation body

   procedure Set_Priority_Defined (N : GS_Node_Id);
   --  Sets for N that for the corresponding entity the priority is already
   --  defined

   ----------------------------------------------------------------------------
   -- Representing Protected Types, Objects and Operations in the Call Graph --
   ----------------------------------------------------------------------------

   --  For each protected operation, we register it in the same way as a task
   --  type (as a non-callable entity for which we store a set of calls issued
   --  from the corresponding protected body, for this we use
   --  A_Protected_Procedure_Body, A_Protected_Function_Body and
   --  A_Protected_Entry_Body node kinds). For each protected object, we
   --  consider a separate "instances" of all the corresponding protected
   --  operations (and use for this A_Protected_Procedure,
   --  A_Protected_Function and A_Protected_Entry node kinds) sharing the same
   --  bodies (and, therefore, the same call chains)

   --  For each protected operation, the SLOC field points to the location of
   --  the declaration of a protected object the operation belongs to, and
   --  Node_Field_1 should be the link to the corresponding
   --  A_Protected_..._Body node.

   --  That is, if a protected type P_T has a protected operation P_Op, and
   --  if we have two different objects of the protected type P_T P_Obj1 and
   --  P_Obj2, then we consider P_Obj1.P_Op and P_Obj2.P_Op as two different
   --  callable entities

   --  As it is implemented now, if a given protected operation belonging to a
   --  given protected object is not called, it is not represented in the
   --  Call Graph.

   --  ??? The same questions as for tasks:
   --  ??? Dynamically created protected objects
   --  ??? Protected objects being components of other data structures
   --  ??? Protected objects as parameters

   -----------------------------------------
   -- Representing (subprogram) renamings --
   -----------------------------------------

   --  Renaming-as-body is treated as a body of a separate callable entity
   --  (that is, it itself is not considered as a callable entity). This entity
   --  has the same call chain as the renamed entity, but calls to this entity
   --  are not considered as calls to renamed entity.

   --  Renaming-as-declaration is treated as a separate callable entity. It
   --  has the same call chain as the renamed entity, and each call to a
   --  renaming entity is also treated as a call to the renamed entity. That
   --  is a set of all the calls to a renaming entity is a subset of the set of
   --  all the calls to the renamed entity.

   --  For a function renaming-as-declaration, in case if a renamed entity is
   --  an enumeration literal, we consider such a renaming as defining a new
   --  function, this function does not call anything but it may be called by
   --  any other node. The idea is to detect unused renamings.

   --  If the renamed entity cannot be statically determined, ???

   --  What happens if we cannot detect the renamed entity because of some
   --  reason???

   -------------------
   -- Add_From_List --
   -------------------

   procedure Add_From_List (To : in out Link_Id; From : Link_Id) is
      Next_Link : Link_Id := From;
      Next_Node : GS_Node_Id;
   begin

      while Present (Next_Link) loop
         Next_Node := Node (Next_Link);
         Add_To_List (To, Next_Node);
         Next_Link := Next (Next_Link);
      end loop;

   end Add_From_List;

   -----------------
   -- Add_To_List --
   -----------------

   procedure Add_To_List
     (To         : in out Link_Id;
      N          :        GS_Node_Id;
      Already_In :    out Boolean)
   is
      Next_Link : Link_Id;
      Add_After : Link_Id;
   begin

      Already_In := False;

      if No (To) then
         --  Empty list
         Entity_Lists.Append
           ((Entity => N,
             Next   => No_Link));

         To := Entity_Lists.Last;
      else
         --  Here we have to check that we have not already stored this node

         Next_Link := To;

         while Present (Next_Link) loop

            if Entity_Lists.Table (Next_Link).Entity = N then
               Already_In := True;
               exit;
            end if;

            Add_After := Next_Link;
            Next_Link := Next (Next_Link);

         end loop;

         if not Already_In then
            Entity_Lists.Append
              ((Entity => N,
                Next   => No_Link));

            Entity_Lists.Table (Add_After).Next := Entity_Lists.Last;
         end if;

      end if;

   end Add_To_List;

   procedure Add_To_List (To : in out Link_Id; N : GS_Node_Id) is
      B : Boolean;
   begin
      Add_To_List (To, N, B);
   end Add_To_List;

   -------------------
   -- Body_Analyzed --
   -------------------

   function Body_Analyzed (N : GS_Node_Id) return Boolean is
   begin
      return GS_Nodes.Table (N).Body_Analyzed;
   end Body_Analyzed;

   ------------------------
   -- Call_To_Unknown_SE --
   ------------------------

   function Call_To_Unknown_SE (Scope : Scope_Id) return Boolean is
   begin
      return GS_Nodes.Table (Scope).Call_To_Unknown_SE;
   end Call_To_Unknown_SE;

   -------------------------------------
   -- Complete_Information_Extraction --
   -------------------------------------

   function Complete_Information_Extraction (N : GS_Node_Id) return Boolean is
      pragma Unreferenced (N);
   begin
      pragma Assert (False);
      --  Placeholder at the moment
      return False;
   end Complete_Information_Extraction;

   -------------------------------
   -- Contains_Dispatching_Call --
   -------------------------------

   function Contains_Dispatching_Call (N : GS_Node_Id) return Boolean is
   begin
      return GS_Nodes.Table (N).Contains_Dispatching_Call;
   end Contains_Dispatching_Call;

   ---------------------------
   -- Contains_Dynamic_Call --
   ---------------------------

   function Contains_Dynamic_Call (N : GS_Node_Id) return Boolean is
   begin
      return GS_Nodes.Table (N).Contains_Dynamic_Call;
   end Contains_Dynamic_Call;

   ---------------
   -- Copy_List --
   ---------------

   procedure Copy_List (To : in out Link_Id; From : Link_Id) is
      Next_Link : Link_Id := From;
      Next_Node : GS_Node_Id;
   begin
      To := No_Link;

      while Present (Next_Link) loop
         Next_Node := Node (Next_Link);

         Entity_Lists.Append ((Next_Node, No_Link));

         if No (To) then
            To := Entity_Lists.Last;
         else
            Entity_Lists.Table (Entity_Lists.Last - 1).Next :=
              Entity_Lists.Last;
         end if;

         Next_Link := Next (Next_Link);

      end loop;

   end Copy_List;

   --------------------------------
   -- Correct_Side_Effect_Status --
   --------------------------------

   procedure Correct_Side_Effect_Status
     (For_Node  : GS_Node_Id;
      From_Node : GS_Node_Id)
   is
      For_SE  : constant Side_Effect_Statuses := Side_Effect_Status (For_Node);
      From_SE : constant Side_Effect_Statuses :=
        Side_Effect_Status (From_Node);
   begin

      if For_SE < Global_Side_Effect then

         if For_SE < From_SE
           and then
            not (From_SE = Local_Side_Effect
              and then
                 Local_Side_Effect_Level (From_Node) >= Scope_Level (For_Node))
         then
            Set_Side_Effect_Status
              (Of_Scope => For_Node,
               To       => From_SE);
            --  This sets the side effect cause SLOC for For_Node to
            --  Nil_String_Loc

         elsif For_SE = From_SE
            and then
               For_SE = Local_Side_Effect
            and then
               Local_Side_Effect_Level (For_Node) >
               Local_Side_Effect_Level (From_Node)
         then
            Set_Side_Effect_Status
              (Of_Scope => For_Node,
               To       => From_SE);
            --  This sets the side effect cause SLOC for For_Node to
            --  Nil_String_Loc

            Set_Local_Side_Effect_Level
              (For_Node,
               Local_Side_Effect_Level (From_Node));
         end if;

         if Side_Effect_Status (For_Node) = Global_Side_Effect then
            Set_Side_Effect_Defined (For_Node);
            Set_Call_To_Unknown_SE (For_Node, False);
         end if;

      end if;

   end Correct_Side_Effect_Status;

   ------------------------
   -- Corresponding_Node --
   ------------------------

   function Corresponding_Node (El : Asis.Element) return GS_Node_Id is
      Result : GS_Node_Id := Find_Node (El);
   begin

      if No (Result) and then not Is_Nil (El) then
         Result := Register_Node (El);
      end if;

      return Result;
   end Corresponding_Node;

   -------------------------------------
   -- Corresponding_Protected_Op_Node --
   -------------------------------------

   function Corresponding_Protected_Op_Node
     (Protected_Obj : Asis.Element;
      Protected_Op  : GS_Node_Id)
      return GS_Node_Id
   is
      Result          : GS_Node_Id := Find_Node (Protected_Obj, Protected_Op);
      Res_Kind        : GS_Node_Kinds;
      Encl_Scope_El   : Asis.Element;
      Encl_Scope_Node : Scope_Id;
   begin

      if No (Result) then

         case GS_Node_Kind (Protected_Op) is
            when A_Protected_Procedure_Body =>
               Res_Kind := A_Protected_Procedure;
            when A_Protected_Function_Body =>
               Res_Kind := A_Protected_Function;
            when A_Protected_Entry_Body =>
               Res_Kind := A_Protected_Entry;
            when others =>
               pragma Assert (False);
               Res_Kind := Not_A_Node;
         end case;

         Result := Register_Node (Protected_Obj, Node_Kind => Res_Kind);
         Set_Protected_Op (Result, Protected_Op);

         Encl_Scope_El := Enclosing_Element (Protected_Obj);

         while not Is_Scope (Encl_Scope_El) loop
            Encl_Scope_El := Enclosing_Element (Encl_Scope_El);
         end loop;

         Encl_Scope_Node := Corresponding_Node (Encl_Scope_El);
         Set_Enclosing_Scope (Result, Encl_Scope_Node);  --  ?????
      end if;

      return Result;
   end Corresponding_Protected_Op_Node;

   -----------------------------
   -- Corresponding_Task_Node --
   -----------------------------

   function Corresponding_Task_Node
     (El            : Asis.Element;
      Single_Task   : Boolean    := True;
      Task_Def_Node : GS_Node_Id := No_GS_Node)
      return          GS_Node_Id
   is
      Result          : GS_Node_Id := Find_Node (El);
      Definition_Node : GS_Node_Id := Task_Def_Node;
   begin

      if No (Result) then
         Result := Register_Node (El, Node_Kind => A_Task_Object);

         if Single_Task then
            Definition_Node :=
              Corresponding_Node (Object_Declaration_View (El));
         end if;

         Set_Task_Definition (Result, Definition_Node);

      end if;

      return Result;
   end Corresponding_Task_Node;

   -------------------
   -- Current_Scope --
   -------------------

   function Current_Scope return Scope_Id is
   begin

      if Current_Scopes.Last >= Current_Scopes.First then
         return Current_Scopes.Table (Current_Scopes.Last);
      else
         raise Scope_Stack_Error;
      end if;

   end Current_Scope;

   -------------------------
   -- Define_GS_Node_Kind --
   -------------------------

   function Define_GS_Node_Kind (El : Asis.Element) return GS_Node_Kinds is
      Result    :          GS_Node_Kinds      := Not_A_Node;
      Encl_Elem : constant Asis.Element       := Enclosing_Element (El);
      Encl_Kind : constant Flat_Element_Kinds := Flat_Element_Kind (Encl_Elem);
   begin

      --  A partial implementation only!

      case Flat_Element_Kind (El) is
         when Not_An_Element =>
            null;

         when A_Procedure_Declaration =>

            if Encl_Kind = A_Protected_Definition then
               Result := A_Protected_Procedure_Body;
            else
               Result := A_Procedure;
            end if;

         when A_Procedure_Body_Declaration |
              A_Procedure_Body_Stub        =>
            Result := A_Procedure;

         when A_Function_Declaration =>

            if Encl_Kind = A_Protected_Definition then
               Result := A_Protected_Function_Body;
            else
               Result := A_Function;
            end if;

         when A_Function_Body_Declaration |
              A_Function_Body_Stub        =>
            Result := A_Function;

         when A_Package_Declaration      |
              A_Package_Body_Declaration =>

            if Is_Nil (Encl_Elem) then
               Result := A_Library_Package;
            elsif Declaration_Kind (Encl_Elem) =
                  A_Package_Instantiation
                and then
                  Is_Nil (Enclosing_Element (Encl_Elem))
            then
               Result := An_Extended_Library_Package_Instantiation;
            else
               pragma Assert (False);
               null;
            end if;

         when A_Package_Instantiation =>

            if Is_Nil (Encl_Elem) then
               Result := A_library_Package_Instantiation;
            else
               pragma Assert (False);
               null;
            end if;

         when A_Single_Task_Declaration =>
            Result := A_Task_Object;
         when An_Entry_Declaration =>
            Result := A_Protected_Entry_Body;

         when A_Procedure_Renaming_Declaration |
              A_Function_Renaming_Declaration  =>

               if Flat_Element_Kind (El) = A_Function_Renaming_Declaration then
                  Result := A_Function;
               else
                  Result := A_Procedure;
               end if;

         when A_Protected_Definition =>
            Result := A_Protected_Definition;

         when A_Task_Definition =>
            Result := A_Task_Definition;

         when others =>
            Ada.Wide_Text_IO.Put_Line (Debug_Image (El));

            if Is_Text_Available (El) then
               Ada.Wide_Text_IO.Put_Line (Element_Image (El));
            end if;

            pragma Assert (False);
            null;
      end case;

      return Result;
   end Define_GS_Node_Kind;

   ------------------------
   -- Define_Side_Effect --
   ------------------------

   procedure Define_Side_Effect
     (Element              : Asis.Element;
      New_SE_Status        : out Side_Effect_Statuses;
      Is_Unresolved_Call   : out Boolean;
      Change_Data_At_Level : out Scope_Levels)
   is
      Arg_Kind : constant Flat_Element_Kinds := Flat_Element_Kind (Element);

      Changed_El           : Asis.Element := Nil_Element;
      Enclosing_Scope_El   : Asis.Element;
      Enclosing_Scope_Node : GS_Node_Id;

      Called_Entity_El   : Asis.Element;
      Called_Entity_Node : GS_Node_Id;
   begin

      New_SE_Status        := No_Side_Effect;
      Is_Unresolved_Call   := False;
      Change_Data_At_Level := Unknown_Scope_Level;

      case Arg_Kind is
         when An_Assignment_Statement =>
            Changed_El := Changed_Element (Assignment_Variable_Name (Element));
         when A_Procedure_Call_Statement |
              A_Function_Call            =>
            --  What about entry calls???

            if Arg_Kind = A_Function_Call then
               Called_Entity_El := Corresponding_Called_Function (Element);
            else
               Called_Entity_El := Corresponding_Called_Entity (Element);
            end if;

            if Is_Nil (Called_Entity_El) then

               if Is_Call_To_Predefined_Operation (Element)
                 or else
                  Is_Call_To_Attribute_Subprogram (Element)
               then
                  --  No side effect!
                  return;
               end if;

               --  Dynamic call, so
               Is_Unresolved_Call := True;
            else

               if Declaration_Kind (Called_Entity_El) =
                  An_Enumeration_Literal_Specification
               then
                  --  No side effect!
                  return;
               end if;

               Called_Entity_El := Corresponding_Element (Called_Entity_El);

               if Is_Predefined_Operation_Renaming (Called_Entity_El) then
                  --  No side effect!
                  return;
               end if;

               Called_Entity_Node := Corresponding_Node (Called_Entity_El);

               if Side_Effect_Defined (Called_Entity_Node) then
                  --  If for the currrent entity side effect is not defined,
                  --  we can not make any conclusion...

                  New_SE_Status := Side_Effect_Status (Called_Entity_Node);

                  if New_SE_Status = Local_Side_Effect then
                     Change_Data_At_Level :=
                       Local_Side_Effect_Level (Called_Entity_Node);
                  end if;

               end if;

            end if;

            --  We should jump out of this routine here not to get to the
            --  assignment-specific processing
            return;

         when others =>
            pragma Assert (False);
            null;
      end case;

      if Is_Nil (Changed_El) then
         --  This means that we have a dynamic situation here:
         New_SE_Status := Potentially_Global_Side_Effect;
      else
         Enclosing_Scope_El   := Enclosing_Scope (Changed_El);
         Enclosing_Scope_Node := Find_Node (Enclosing_Scope_El);

         if No (Enclosing_Scope_Node)
            --  This means that Enclosing_Scope_El is in a global scope that
            --  has not been processed yet
           or else
            Is_Global_Scope (Enclosing_Scope_Node)
         then
            New_SE_Status := Global_Side_Effect;
         elsif Enclosing_Scope_Node = Current_Scope then
            New_SE_Status := No_Side_Effect;
         else
            --  The only possibility is that Changed_El is in some enclosing
            --  global scope
            New_SE_Status        := Local_Side_Effect;
            Change_Data_At_Level := Scope_Level (Enclosing_Scope_Node);
         end if;

      end if;

   end Define_Side_Effect;

   ------------------------------------
   -- Enclosing_Protected_Definition --
   ------------------------------------

   function Enclosing_Protected_Definition
     (For_PO : GS_Node_Id)
      return   GS_Node_Id
   is
   begin
      pragma Assert
        (GS_Node_Kind (For_PO) in
          A_Protected_Procedure_Body .. A_Protected_Entry_Body);

      return GS_Nodes.Table (For_PO).Node_Field_1;
   end Enclosing_Protected_Definition;

   ---------------------
   -- Enclosing_Scope --
   ---------------------

   function Enclosing_Scope (N : GS_Node_Id) return Scope_Id is
   begin
      return GS_Nodes.Table (N).Enclosing_Scope;
   end Enclosing_Scope;

   ---------------------
   -- Enclosed_Source --
   ---------------------

   function Enclosed_Source (N : GS_Node_Id) return SF_Id is
   begin
      return GS_Nodes.Table (N).SF;
   end Enclosed_Source;

   ---------------
   -- Find_Node --
   ---------------

   function Find_Node
     (El           : Asis.Element;
      Protected_Op : GS_Node_Id := No_GS_Node)
      return         GS_Node_Id
   is
      Result : GS_Node_Id := Hash_Table (Hash (El));
   begin

      if Is_Nil (El) then
         return No_GS_Node;
      end if;

      while Present (Result) loop

         if Is_Equal (Result, El, Protected_Op) then
            exit;
         end if;

         Result := GS_Nodes.Table (Result).Hash_Link;
      end loop;

      return Result;
   end Find_Node;

   -------------------------
   -- First_Called_Entity --
   -------------------------

   function First_Called_Entity (N : GS_Node_Id) return Link_Id is
   begin
      return GS_Nodes.Table (N).All_Calls_Chain;
   end First_Called_Entity;

   --------------------------
   -- First_Calling_Entity --
   --------------------------

   function First_Calling_Entity (N : GS_Node_Id) return Link_Id is
   begin
      return GS_Nodes.Table (N).All_Callers_Chain;
   end First_Calling_Entity;

   ----------------------------------
   -- First_Directly_Called_Entity --
   ----------------------------------

   function First_Directly_Called_Entity (N : GS_Node_Id) return Link_Id is
   begin
      return GS_Nodes.Table (N).Calls_Chain;
   end First_Directly_Called_Entity;

   -----------------------------------
   -- First_Directly_Calling_Entity --
   -----------------------------------

   function First_Directly_Calling_Entity (N : GS_Node_Id) return Link_Id is
   begin
      return GS_Nodes.Table (N).Callers_Chain;
   end First_Directly_Calling_Entity;

   --------------------------
   -- First_Enclosed_Scope --
   --------------------------

   function First_Enclosed_Scope (N : GS_Node_Id) return Scope_Id is
   pragma Unreferenced (N);
   begin
      pragma Assert (False);
      --  Placeholder at the moment
      return No_Scope;
   end First_Enclosed_Scope;

   ----------------------
   -- Get_Protected_Op --
   ----------------------

   function Get_Protected_Op (For_Task : GS_Node_Id) return GS_Node_Id is
   begin
      pragma Assert (GS_Node_Kind (For_Task) in
                       A_Protected_Procedure .. A_Protected_Entry);
      return GS_Nodes.Table (For_Task).Node_Field_1;
   end Get_Protected_Op;

   -------------------------
   -- Get_Task_Definition --
   -------------------------

   function Get_Task_Definition (For_Task : GS_Node_Id) return GS_Node_Id is
   begin
      pragma Assert (GS_Node_Kind (For_Task) = A_Task_Object);
      return GS_Nodes.Table (For_Task).Node_Field_1;
   end Get_Task_Definition;

   ------------------
   -- GS_Node_Kind --
   ------------------

   function GS_Node_Kind (N : GS_Node_Id) return GS_Node_Kinds is
   begin
      return GS_Nodes.Table (N).Node_Kind;
   end GS_Node_Kind;

   ---------------------------
   -- GS_Node_Renaming_Kind --
   ---------------------------

   function GS_Node_Renaming_Kind (N : GS_Node_Id) return Renaming_Kinds is
   begin
      return GS_Nodes.Table (N).Renaming_Kind;
   end GS_Node_Renaming_Kind;

   ----------
   -- Hash --
   ----------

   function Hash (El : Asis.Element) return Hash_Index_Type is
   begin
      return Asis.Elements.Hash (El) mod Hash_Num;
   end Hash;

   --------------------------
   -- Has_Dynamic_Priority --
   --------------------------

   function Has_Dynamic_Priority (N : GS_Node_Id) return Boolean is
   begin
      return
         Priority_Defined (N)
       and then
         GS_Nodes.Table (N).Has_Dynamic_Priority;
   end Has_Dynamic_Priority;

   -------------
   -- In_List --
   -------------

   function In_List (N : GS_Node_Id; L : Link_Id) return Boolean is
      Result    : Boolean := False;
      Next_Link : Link_Id := L;
   begin

      while Present (Next_Link) loop
         if Entity_Lists.Table (Next_Link).Entity = N then
            Result := True;
            exit;
         end if;

         Next_Link := Next (Next_Link);

      end loop;

      return Result;
   end In_List;

   ----------------
   -- Initialize --
   ----------------

   procedure Initialize is
   begin
      Current_Scopes.Init;
      GS_Nodes.Init;
      Hash_Table := (others => No_GS_Node);

      --  Locating the node representing the evironment task:
      GS_Nodes.Append (Environment_Task_Node_Rec);
      Environment_Task_Node := GS_Nodes.Last;

      Set_Current_Scope (GS_Nodes.Last);

   end Initialize;

   --------------
   -- Is_Equal --
   --------------

   function Is_Equal
     (GS_Node      : GS_Node_Id;
      El           : Asis.Element;
      Protected_Op : GS_Node_Id := No_GS_Node)
      return         Boolean
   is
      N_Kind  :  constant GS_Node_Kinds := GS_Node_Kind (GS_Node);
      Op_Kind :  GS_Node_Kinds;
      Result  : Boolean := False;
   begin

      if Present (GS_Node) then

         if Present (Protected_Op) then
            Op_Kind := GS_Node_Kind (Protected_Op);

            if (N_Kind = A_Protected_Procedure
               and then
                Op_Kind = A_Protected_Procedure_Body)
              or else
               (N_Kind = A_Protected_Function
               and then
                Op_Kind = A_Protected_Function_Body)
              or else
               (N_Kind = A_Protected_Entry
               and then
                Op_Kind = A_Protected_Entry_Body)
            then
               Result :=
                  Protected_Op = Get_Protected_Op (GS_Node)
                 and then
                  Get_String (GS_Nodes.Table (GS_Node).SLOC) =
                  Build_GNAT_Location (El);
            end if;

         elsif Get_String (GS_Nodes.Table (GS_Node).SLOC) =
               Build_GNAT_Location (El)
         then
            Result :=
               not (N_Kind = A_Protected_Definition
                  or else
                    N_Kind = A_Protected_Procedure_Body
                  or else
                    N_Kind = A_Protected_Function_Body
                  or else
                    N_Kind = A_Protected_Entry_Body)
             or else
               N_Kind = Define_GS_Node_Kind (El);
         end if;

      end if;

      return Result;
   end Is_Equal;

   ---------------------
   -- Is_Global_Scope --
   ---------------------

   function Is_Global_Scope (Scope : Scope_Id) return Boolean is
   begin

      if Scope = Environment_Task_Node then
         return True;
      else
         return Is_Scope (Scope)
              and then
                GS_Nodes.Table (Scope).Scope_Level = 0;
      end if;

   end Is_Global_Scope;

   --------------------
   -- Is_Local_Scope --
   --------------------

   function Is_Local_Scope (N : GS_Node_Id) return Boolean is
   begin
      return Is_Scope (N) and then GS_Nodes.Table (N).Scope_Level > 0;
   end Is_Local_Scope;

   -----------------
   -- Is_RTL_Node --
   -----------------

   function Is_RTL_Node (N : GS_Node_Id) return Boolean is
   begin
      return GS_Nodes.Table (N).Is_RTL_Node;
   end Is_RTL_Node;

   --------------
   -- Is_Scope --
   --------------

   function Is_Scope (N : GS_Node_Id) return Boolean is
   begin
      return GS_Nodes.Table (N).Is_Scope;
   end Is_Scope;

   -------------------
   -- Is_Subprogram --
   -------------------

   function Is_Subprogram (N : GS_Node_Id) return Boolean is
   begin
      return GS_Nodes.Table (N).Node_Kind in
                A_Procedure .. A_Function_Instantiation;
   end Is_Subprogram;

   -----------------------------
   -- Local_Side_Effect_Level --
   -----------------------------

   function Local_Side_Effect_Level
     (Of_Scope : Scope_Id)
      return     Scope_Levels
   is
   begin
      return GS_Nodes.Table (Of_Scope).Local_Side_Effect_Level;
   end Local_Side_Effect_Level;

   --------------
   -- Location --
   --------------

   function Location (N : GS_Node_Id) return String_Loc is
   begin
      return GS_Nodes.Table (N).SLOC;
   end Location;

   ----------
   -- Next --
   ----------

   function Next (L : Link_Id) return Link_Id is
   begin
      pragma Assert (Present (L));
      return Entity_Lists.Table (L).Next;
   end Next;

   ----------------
   -- Next_Scope --
   ----------------

   function Next_Scope (S : Scope_Id) return Scope_Id is
   pragma Unreferenced (S);
   begin
      pragma Assert (False);
      --  Placeholder at the moment
      return No_Scope;
   end Next_Scope;

   --------
   -- No --
   --------

   function No (N : GS_Node_Id) return Boolean is
   begin
      return N not in First_GS_Node .. GS_Nodes.Last;
   end No;

   function No (L : Link_Id) return Boolean is
   begin
      return L not in First_Link_Id .. Entity_Lists.Last;
   end No;

   ----------
   -- Node --
   ----------

   function Node (L : Link_Id) return GS_Node_Id is
   begin

      if Present (L) then
         return Entity_Lists.Table (L).Entity;
      else
         return No_GS_Node;
      end if;

   end Node;

   -------------------
   -- Node_Priority --
   -------------------

   function Node_Priority (N : GS_Node_Id) return Any_Priority is
   begin
      return GS_Nodes.Table (N).Node_Priority;
   end Node_Priority;

   -------------
   -- Present --
   -------------

   function Present (N : GS_Node_Id) return Boolean is
   begin
      return N in First_GS_Node .. GS_Nodes.Last;
   end Present;

   function Present (L : Link_Id) return Boolean is
   begin
      return L in First_Link_Id .. Entity_Lists.Last;
   end Present;

   ----------------------------
   -- Print_Global_Structure --
   ----------------------------

   procedure Print_Global_Structure is
   begin
      Put_Line ("*** NODES TABLE ***");

      for J in First_GS_Node .. GS_Nodes.Last loop
         Print_Node (J);
      end loop;

      Print_Node_Lists_Table;

   end Print_Global_Structure;

   ----------------
   -- Print_Node --
   ----------------

   procedure Print_Node (N : GS_Node_Id) is
   begin
      Put_Line ("Node_Id =" & N'Img);

      Put      (Ident_String);
      Put_Line ("Node_Kind       = " &  GS_Nodes.Table (N).Node_Kind'Img);

      Put      (Ident_String);
      Put_Line ("SF              =" & GS_Nodes.Table (N).SF'Img);

      Put      (Ident_String);
      Put_Line ("SLOC            = " & Get_String (GS_Nodes.Table (N).SLOC));

      Put      (Ident_String);
      Put      ("Enclosing_Scope =" & Enclosing_Scope (N)'Img);

      if Enclosing_Scope (N) = No_Scope then
         Put (" (****** INDEFINED!!!)");
      end if;

      New_Line;
      --  ????

      if Is_Scope (N) then
         Put      (Ident_String & Ident_String);
         Put_Line ("Scope level        =" &
                    Scope_Level (N)'Img);

      end if;

      Put (Ident_String);
      Put ("Side Effect Status = " & GS_Nodes.Table (N).SE_Status'Img);

      if not (GS_Nodes.Table (N).SE_Status = No_Side_Effect
            and then
              GS_Nodes.Table (N).Side_Effect_Defined)
      then

         if GS_Nodes.Table (N).SE_Status = Local_Side_Effect then
            Put (" (" & GS_Nodes.Table (N).Local_Side_Effect_Level'Img &
                 ')');
         end if;

         New_Line;

         if GS_Nodes.Table (N).SE_SLOC /= Nil_String_Loc then
            Put      (Ident_String & Ident_String);
            Put_Line ("Side Effec cause location is " &
                      Get_String (GS_Nodes.Table (N).SE_SLOC));
         end if;

         Put      (Ident_String & Ident_String);
         Put_Line ("Side_Effect_Defined = " &
                   GS_Nodes.Table (N).Side_Effect_Defined'Img);

         Put      (Ident_String & Ident_String);
         Put_Line ("Call_To_Unknown_SE = " &
                   GS_Nodes.Table (N).Call_To_Unknown_SE'Img);
      else
         New_Line;
      end if;

      Put      (Ident_String);
      Put_Line ("Body_Analyzed   = " & GS_Nodes.Table (N).Body_Analyzed'Img);

      Put      (Ident_String);
      Put_Line ("Hash_Link       =" & GS_Nodes.Table (N).Hash_Link'Img);

      case GS_Nodes.Table (N).Node_Kind is

         when A_Task_Object =>
            Put      (Ident_String);
            Put      (Ident_String);
            Put_Line ("Corresponding task def  =" &
                      Get_Task_Definition (N)'Img);
         when A_Protected_Procedure .. A_Protected_Entry =>
            Put      (Ident_String);
            Put      (Ident_String);
            Put_Line ("Executable protected operation =" &
                       Get_Protected_Op (N)'Img);
         when A_Protected_Procedure_Body .. A_Protected_Entry_Body =>
            Put      (Ident_String);
            Put      (Ident_String);
            Put_Line ("Enclosing Protected Definition =" &
                       Enclosing_Protected_Definition (N)'Img);
         when others =>
            null;
      end case;

      if GS_Node_Renaming_Kind (N) /= Not_A_Renamimg then
         Put      (Ident_String);
         Put_Line ("Renaming_Kind   = " & GS_Node_Renaming_Kind (N)'Img);

         if GS_Node_Renaming_Kind (N) /= Enum_Literal_Renaming then
            Put      (Ident_String);
            Put      (Ident_String);
            Put_Line ("Renamed entity =" & Renamed_Entity (N)'Img);
         end if;
      end if;

      Put (Ident_String);
      Put ("Priority ");

      if Priority_Defined (N) then

         if Has_Dynamic_Priority (N) then
            Put_Line ("dynamic");
         else
            Put_Line ("=" & Node_Priority (N)'Img);
         end if;

      else
         Put_Line ("**INDEFINED**");
      end if;

      Put      (Ident_String);
      Put_Line ("Is_Used         = " & GS_Nodes.Table (N).Is_Used'Img);

      Put (Ident_String & "Direct calls:");
      Print_Node_List (GS_Nodes.Table (N).Calls_Chain);

      Put (Ident_String & "Direct callers:");
      Print_Node_List (GS_Nodes.Table (N).Callers_Chain);

      Put (Ident_String & "All calls:");
      Print_Node_List (GS_Nodes.Table (N).All_Calls_Chain);

      Put (Ident_String & "All callers:");
      Print_Node_List (GS_Nodes.Table (N).All_Callers_Chain);

      if GS_Nodes.Table (N).All_CG_Details_Set then
         Put      (Ident_String);
         Put_Line ("All details set");

         if not GS_Nodes.Table (N).All_Calls_Set then
            Put      (Ident_String);
            Put_Line ("!!! All calls are NOT set");
         end if;

         if not GS_Nodes.Table (N).All_Callers_Set then
            Put      (Ident_String);
            Put_Line ("!!! All callers are NOT set");
         end if;

      else

         Put (Ident_String);

         if GS_Nodes.Table (N).All_Calls_Set then
            Put_Line ("All calls set");
         else
            Put_Line ("All calls are not set");
         end if;

         Put (Ident_String);

         if GS_Nodes.Table (N).All_Callers_Set then
            Put_Line ("All callers set");
         else
            Put_Line ("All callers are not set");
         end if;

      end if;

      New_Line;
   end Print_Node;

   ---------------------
   -- Print_Node_List --
   ---------------------

   procedure Print_Node_List (L : Link_Id; Ident : Natural := 0) is
      Next_El : Link_Id := L;
   begin

      for J in 1 .. Ident loop
         Put (Ident_String);
      end loop;

      Put (" (Link_Id =" & L'Img & ")");

      if Present (Next_El) then

         while Present (Next_El) loop
            Put (Entity_Lists.Table (Next_El).Entity'Img);
            Next_El := Next (Next_El);
         end loop;

      else
         Put (" ...nothing...");
      end if;

      New_Line;

   end Print_Node_List;

   ----------------------------
   -- Print_Node_Lists_Table --
   ----------------------------

   procedure Print_Node_Lists_Table is
   begin
      Put_Line ("*** Low-level node list info ***");

      for J in First_Link_Id .. Entity_Lists.Last loop
         Put_Line ("Link_id =" & J'Img);
         Put (Ident_String);
         Put_Line ("Entity =" & Entity_Lists.Table (J).Entity'Img);
         Put (Ident_String);
         Put_Line ("Next   =" & Entity_Lists.Table (J).Next'Img);
         New_Line;
      end loop;

   end Print_Node_Lists_Table;

   ----------------------
   -- Priority_Defined --
   ----------------------

   function Priority_Defined (N : GS_Node_Id) return Boolean is
   begin
      return GS_Nodes.Table (N).Priority_Defined;
   end Priority_Defined;

   ---------------------
   -- Register_Entity --
   ---------------------

   function Register_Entity
     (El               : Asis.Element;
      Fix_Protected_Op : Boolean := False)
      return             GS_Node_Id
   is
      Corresponding_El : constant Asis.Element := Corresponding_Element (El);
      Result           : GS_Node_Id            := Find_Node (Corresponding_El);

      Separate_Body : Asis.Element;
      Tmp_SF        : SF_Id;

      Tmp_Node : GS_Node_Id;

   begin
      if No (Result) then
         Result := Register_Node (Corresponding_El);
      end if;

      if Is_Body (El)
          or else
            Is_Renaming_As_Body (El)
      then
         Set_Body_Analyzed (Result, True);
      end if;

      if Declaration_Kind (El) in
         A_Procedure_Body_Stub .. A_Protected_Body_Stub
      then

         Separate_Body := Corresponding_Subunit (El);

         if Is_Nil (Separate_Body) then
            Error ("cannot locate the proper body for subunit ???");
         else
            Tmp_SF := File_Find (Separate_Body);

            if not Present (Tmp_SF) then
               Tmp_SF :=
                 Add_Needed_Source
                   (Normalize_Pathname
                      (To_String
                         (Asis.Compilation_Units.Text_Name
                            (Asis.Elements.Enclosing_Compilation_Unit
                               (Separate_Body)))));
               Gnatcheck.Diagnoses.Add_Line_To_Mapping_Table;
            end if;

         end if;
      end if;

      if Fix_Protected_Op
        and then
           (Declaration_Kind (El) in
              A_Procedure_Declaration .. A_Function_Declaration
           or else
            Declaration_Kind (El) = An_Entry_Declaration)
        and then
         GS_Node_Kind (Result) in
            A_Protected_Procedure_Body .. A_Protected_Entry_Body
        and then
         No (Enclosing_Protected_Definition (Result))
      then
         Tmp_Node := Find_Node (Enclosing_Element (El));
         pragma Assert (Present (Tmp_Node));
         Set_Enclosing_Protected_Definition (Result, Tmp_Node);
      end if;

      return Result;
   end Register_Entity;

   procedure Register_Entity
     (El               : Asis.Element;
      Fix_Protected_Op : Boolean := False)
   is
      Tmp : GS_Node_Id := Register_Entity (El, Fix_Protected_Op);
      pragma Warnings (Off, Tmp);
   begin
      null;
   end Register_Entity;

   -------------------
   -- Register_Node --
   -------------------

   function Register_Node
     (El         : Asis.Element;
      Node_Kind  : GS_Node_Kinds := Not_A_Node)
      return       GS_Node_Id
   is
      Hash_Value       : constant Hash_Index_Type := Hash (El);
      Last_In_Chain    : GS_Node_Id               := Hash_Table (Hash_Value);
      Node_Kind_To_Set : GS_Node_Kinds            := Node_Kind;
      Is_Used          : Boolean                  := False;
      Encl_Scope       : Scope_Id                 := No_Scope;
      Body_Analyzed    : Boolean                  := not Needs_Completion (El);
      New_Node         : GS_Node_Record;
      Node_SF          : SF_Id                    := No_SF_Id;
      Result           : GS_Node_Id;
      Is_RTL_Node      : constant Boolean        :=
         Unit_Origin (Enclosing_Compilation_Unit (El)) /= An_Application_Unit;
   begin

      if Node_Kind_To_Set = Not_A_Node then
         Node_Kind_To_Set := Define_GS_Node_Kind (El);
      end if;

      if Node_Kind_To_Set in A_Procedure .. A_Function_Instantiation
        and then
         Is_Nil (Enclosing_Element (El))
        and then
         Can_Be_Main_Program (Enclosing_Compilation_Unit (El))
      then
         Is_Used := True;
      end if;

      if Node_Kind_To_Set = A_Protected_Definition then
         --  We can get into creation of the node for a protected definition
         --  only in the regular traversing
         Encl_Scope     := Current_Scope;

         --  We do not care about analyzing bodies for protected definitions -
         --  the check that is made for protected operations is enough
         Body_Analyzed := True;
      end if;

      if not Process_RTL_Units
       and then
         Is_RTL_Node
      then
         Is_Used       := True;
         Encl_Scope    := Environment_Task_Node;
         Body_Analyzed := True;
      else
         Node_SF := File_Find (El);

         if not Present (Node_SF) then
            Node_SF :=
               Add_Needed_Source
                 (Normalize_Pathname
                    (To_String
                       (Asis.Compilation_Units.Text_Name
                          (Asis.Elements.Enclosing_Compilation_Unit (El)))));

            Gnatcheck.Diagnoses.Add_Line_To_Mapping_Table;
         end if;

      end if;

      New_Node :=
        (Node_Kind                 => Node_Kind_To_Set,
         SF                        => Node_SF,
         SLOC                      => Build_GNAT_Location (El),
         Enclosing_Scope           => Encl_Scope,                    --  ????
         Is_Scope                  => Is_Scope (El),
         Scope_Level               => Unknown_Scope_Level,
         Local_Side_Effect_Level   => Unknown_Scope_Level,
         SE_Status                 => Unknown,
         SE_SLOC                   => Nil_String_Loc,
         Side_Effect_Defined       => False,
         Call_To_Unknown_SE        => False,
         Node_Priority             => Default_Priority,
         Has_Dynamic_Priority      => False,
         Priority_Defined          => False,
         Body_Analyzed             => Body_Analyzed,
         Calls_Chain               => No_Link,
         Callers_Chain             => No_Link,
         Hash_Link                 => No_GS_Node,
         Node_Field_1              => No_GS_Node,
         Contains_Dispatching_Call => False,
         Contains_Dynamic_Call     => False,
         Renaming_Kind             => Not_A_Renamimg,
         Is_Dynamic                => False,
         Is_Used                   => Is_Used,
         Is_RTL_Node               => Is_RTL_Node,
         All_Callers_Set           => False,
         All_Calls_Set             => False,
         All_CG_Details_Set        => False,
         All_Callers_Chain         => No_Link,
         All_Calls_Chain           => No_Link);

--      pragma Assert (False
--        or else Flat_Element_Kind (El) = A_Procedure_Declaration
--        or else Flat_Element_Kind (El) = A_Function_Declaration
--        or else ((Flat_Element_Kind (El) = A_Procedure_Body_Declaration
--                or else
--                 Flat_Element_Kind (El) = A_Function_Body_Declaration)
--             and then
--                 Acts_As_Spec (El)));

      pragma Assert
         (Present (New_Node.SF)
         or else
          (not Process_RTL_Units and then Is_RTL_Node));

      if Declaration_Kind (El) in
           A_Procedure_Declaration .. A_Function_Declaration
        and then
         not Needs_Completion (El)
      then
         --  This means that El is a subprogram that do not have the body in
         --  Ada
         New_Node.SE_Status           := No_Side_Effect;
         New_Node.Side_Effect_Defined := True;
      end if;

      if not Process_RTL_Units
       and then
         Is_RTL_Node
      then
         --  We have to continue to set stubs for non-analysable RTL node
         --  fields
         New_Node.SE_Status           := No_Side_Effect;
         New_Node.Side_Effect_Defined := True;
--         New_Node.Priority_Defined    := True;        ???
         New_Node.Priority_Defined    := True;
         New_Node.All_Calls_Set       := True;
      end if;

      GS_Nodes.Append (New_Node);
      Result := GS_Nodes.Last;

      if No (Last_In_Chain) then
         Hash_Table (Hash_Value) := Result;
      else
         while Present (GS_Nodes.Table (Last_In_Chain).Hash_Link) loop
            Last_In_Chain := GS_Nodes.Table (Last_In_Chain).Hash_Link;
         end loop;

         GS_Nodes.Table (Last_In_Chain).Hash_Link := Result;
      end if;

      if not (not Process_RTL_Units
           and then
             Is_RTL_Node)
      then
         Store_Needed_Source (El);
      end if;

      return Result;
   end Register_Node;

   --------------------------
   -- Remove_Current_Scope --
   --------------------------

   procedure Remove_Current_Scope is
   begin

      if Current_Scopes.Last >= Current_Scopes.First then
         Current_Scopes.Decrement_Last;
      else
         raise Scope_Stack_Error;
      end if;

   end Remove_Current_Scope;

   --------------------
   -- Renamed_Entity --
   --------------------

   function Renamed_Entity (N : GS_Node_Id) return GS_Node_Id is
   begin
      --  ??? pragma Assert (???);
      return GS_Nodes.Table (N).Node_Field_1;
   end Renamed_Entity;

   -----------------
   -- Scope_Level --
   -----------------

   function Scope_Level (Scope : Scope_Id) return Scope_Levels is
   begin
      return GS_Nodes.Table (Scope).Scope_Level;
   end Scope_Level;

   -----------------------
   -- Set_Body_Analyzed --
   -----------------------

   procedure Set_Body_Analyzed (N : GS_Node_Id; Val : Boolean) is
   begin
      GS_Nodes.Table (N).Body_Analyzed := Val;
   end Set_Body_Analyzed;

   -----------------------
   -- Set_Current_Scope --
   -----------------------

   procedure Set_Current_Scope (Scope : Asis.Element) is
   begin
      Set_Current_Scope (Corresponding_Node (Corresponding_Element (Scope)));
   end Set_Current_Scope;

   procedure Set_Current_Scope (Scope : GS_Node_Id) is
   begin
      --  ??? Links between scopes ???

      if Scope /= Environment_Task_Node then
         Set_Enclosing_Scope (Scope);
      end if;

      Current_Scopes.Append (Scope);

      if GS_Nodes.Table (Scope).Scope_Level = Unknown_Scope_Level then

         if GS_Node_Kind (Scope) in
            A_Library_Package .. An_Extended_Library_Package_Instantiation
         then
            GS_Nodes.Table (Scope).Scope_Level := Global_Scope_Level;

         elsif GS_Node_Kind (Current_Scopes.Table (Top_Scope_Ind)) in
            A_Library_Package .. An_Extended_Library_Package_Instantiation
         then
            GS_Nodes.Table (Scope).Scope_Level := Current_Scopes.Last - 1;

         else
            GS_Nodes.Table (Scope).Scope_Level := Current_Scopes.Last;
         end if;

      end if;

      if Side_Effect_Status (Scope) = Unknown then
         Set_Side_Effect_Status (Scope, No_Side_Effect);
      end if;

   end Set_Current_Scope;

   ----------------------------
   -- Set_Call_To_Unknown_SE --
   ----------------------------

   procedure Set_Call_To_Unknown_SE
     (Scope : Scope_Id;
      Val   : Boolean := True)
   is
   begin
      GS_Nodes.Table (Scope).Call_To_Unknown_SE := Val;
   end Set_Call_To_Unknown_SE;

   -----------------------------------
   -- Set_Contains_Dispatching_Call --
   -----------------------------------

   procedure Set_Contains_Dispatching_Call (N : GS_Node_Id) is
   begin
      GS_Nodes.Table (N).Contains_Dispatching_Call := True;
   end Set_Contains_Dispatching_Call;

   -------------------------------
   -- Set_Contains_Dynamic_Call --
   -------------------------------

   procedure Set_Contains_Dynamic_Call (N : GS_Node_Id) is
   begin
      GS_Nodes.Table (N).Contains_Dynamic_Call := True;
   end Set_Contains_Dynamic_Call;

   --------------------------
   -- Set_Dynamic_Priority --
   --------------------------

   procedure Set_Dynamic_Priority (N : GS_Node_Id) is
   begin
      GS_Nodes.Table (N).Has_Dynamic_Priority := True;
      Set_Priority_Defined (N);
   end Set_Dynamic_Priority;

   ----------------------------------------
   -- Set_Enclosing_Protected_Definition --
   ----------------------------------------

   procedure Set_Enclosing_Protected_Definition
     (For_PO : GS_Node_Id;
      P_Def  : GS_Node_Id)
   is
   begin
      pragma Assert
        (GS_Node_Kind (For_PO) in
          A_Protected_Procedure_Body .. A_Protected_Entry_Body);

      GS_Nodes.Table (For_PO).Node_Field_1 := P_Def;
   end Set_Enclosing_Protected_Definition;

   -------------------------
   -- Set_Enclosing_Scope --
   -------------------------

   procedure Set_Enclosing_Scope
     (N : GS_Node_Id;
      S : Scope_Id := Current_Scope)
   is
   begin
      GS_Nodes.Table (N).Enclosing_Scope := S;
   end Set_Enclosing_Scope;

   -------------------------------
   -- Set_GS_Node_Renaming_Kind --
   -------------------------------

   procedure Set_GS_Node_Renaming_Kind
     (N   : GS_Node_Id;
      Val : Renaming_Kinds)
   is
   begin
      GS_Nodes.Table (N).Renaming_Kind := Val;
   end Set_GS_Node_Renaming_Kind;

   ---------------------------------
   -- Set_Local_Side_Effect_Level --
   ---------------------------------

   procedure Set_Local_Side_Effect_Level
     (Of_Scope : Scope_Id;
      To_Level : Scope_Levels)
   is
   begin
      GS_Nodes.Table (Of_Scope).Local_Side_Effect_Level := To_Level;
   end Set_Local_Side_Effect_Level;

   ------------------
   -- Set_Priority --
   ------------------

   procedure Set_Priority (N : GS_Node_Id; Val : Any_Priority) is
   begin
      GS_Nodes.Table (N).Node_Priority := Val;
      Set_Priority_Defined (N);
   end Set_Priority;

   --------------------------
   -- Set_Priority_Defined --
   --------------------------

   procedure Set_Priority_Defined (N : GS_Node_Id) is
   begin
      GS_Nodes.Table (N).Priority_Defined := True;
   end Set_Priority_Defined;

   ----------------------
   -- Set_Protected_Op --
   ----------------------

   procedure Set_Protected_Op (For_Obj : GS_Node_Id; Pr_Op : GS_Node_Id) is
   begin
      pragma Assert (GS_Node_Kind (For_Obj) in
                       A_Protected_Procedure .. A_Protected_Entry);
      GS_Nodes.Table (For_Obj).Node_Field_1 := Pr_Op;
   end Set_Protected_Op;

   ------------------------
   -- Set_Renamed_Entity --
   ------------------------

   procedure Set_Renamed_Entity (For_Node : GS_Node_Id; Val : GS_Node_Id) is
   begin
      --  ??? pragma Assert (???);
      GS_Nodes.Table (For_Node).Node_Field_1 := Val;
   end Set_Renamed_Entity;

   -----------------------------
   -- Set_Side_Effect_Defined --
   -----------------------------

   procedure Set_Side_Effect_Defined
     (Of_Scope : Scope_Id;
      Val      : Boolean := True)
   is
   begin
      GS_Nodes.Table (Of_Scope).Side_Effect_Defined := Val;
   end Set_Side_Effect_Defined;

   ----------------------------
   -- Set_Side_Effect_Status --
   ----------------------------

   procedure Set_Side_Effect_Status
     (Of_Scope : Scope_Id;
      To       : Side_Effect_Statuses;
      At_SLOC  : String_Loc := Nil_String_Loc)
   is
   begin
      GS_Nodes.Table (Of_Scope).SE_Status := To;
      GS_Nodes.Table (Of_Scope).SE_SLOC   := At_SLOC;
   end Set_Side_Effect_Status;

   -------------------------
   -- Set_Task_Definition --
   -------------------------

   procedure Set_Task_Definition (T : GS_Node_Id; T_Def : GS_Node_Id) is
   begin
      pragma Assert (GS_Node_Kind (T) = A_Task_Object);
      GS_Nodes.Table (T).Node_Field_1 := T_Def;
   end Set_Task_Definition;

   -----------------------
   -- Side_Effect_Cause --
   -----------------------

   function Side_Effect_Cause (N : GS_Node_Id) return Side_Effect_Causes is
   pragma Unreferenced (N);
   begin
      pragma Assert (False);
      --  Placeholder at the moment
      return No_Side_Effec_Cause;
   end Side_Effect_Cause;

   ---------------------------
   -- Side_Effect_Cause_Loc --
   ---------------------------

   function Side_Effect_Cause_Loc (N : GS_Node_Id) return String_Loc is
   pragma Unreferenced (N);
   begin
      pragma Assert (False);
      --  Placeholder at the moment
      return Nil_String_Loc;
   end Side_Effect_Cause_Loc;

   -------------------------
   -- Side_Effect_Defined --
   -------------------------

   function Side_Effect_Defined (Of_Scope : Scope_Id) return Boolean is
   begin
      return GS_Nodes.Table (Of_Scope).Side_Effect_Defined;
   end Side_Effect_Defined;

   ------------------------
   -- Side_Effect_Status --
   ------------------------

   function Side_Effect_Status
     (Scope : Scope_Id)
      return  Side_Effect_Statuses
   is
   begin
      return GS_Nodes.Table (Scope).SE_Status;
   end Side_Effect_Status;

   ---------------
   -- Store_Arc --
   ---------------

   procedure Store_Arc (Called_Node : GS_Node_Id) is
      Already_Stored : Boolean;
   begin

      --  First, store the Current_Scope --calls--> Called_Node arc:

      Add_To_List
        (To         => GS_Nodes.Table (Current_Scope).Calls_Chain,
         N          => Called_Node,
         Already_In => Already_Stored);

      --  Now, store the Called_Node --is called by--> Current_Scope arc:

      if not Already_Stored then
         Add_To_List
           (To         => GS_Nodes.Table (Called_Node).Callers_Chain,
            N          => Current_Scope);
      end if;

   end Store_Arc;

   procedure Store_Arc (Called_Entity : Asis.Element) is
      Called_Node : constant GS_Node_Id := Corresponding_Node (Called_Entity);
   begin
      pragma Assert (Present (Called_Node));
      Store_Arc (Called_Node);
   end Store_Arc;

   -------------------------
   -- Store_Needed_Source --
   -------------------------

   procedure Store_Needed_Source (El : Asis.Element) is
      Completion_El : Asis.Element;
      SF            : SF_Id;
   begin
      --  If we have N already stored in the Call Graph node table, it means
      --  that for N itself we have already stored the source file - either as
      --  an argument file or as a needed file. It is not enough if El (for
      --  that N has been created and stored in the Call Graph) is a
      --  declaration requiring completion

      if Needs_Completion (El)
        and then
         Unit_Kind (Enclosing_Compilation_Unit (El)) not in
           A_Procedure_Instance .. A_Package_Instance
      then

         Completion_El := Corresponding_Body (El); --  ???

         if Is_Nil (Completion_El) then
            --  Try to compute the name of the corresponding body file

            declare
               Body_Name : constant String := Get_Body_Name (El);
            begin

               if Is_Regular_File (Body_Name) then
                  SF := File_Find (Body_Name);

                  if not Present (SF) then
                     SF := Add_Needed_Source (Body_Name);
                     Gnatcheck.Diagnoses.Add_Line_To_Mapping_Table;
                  end if;
               else
                  Error
                    ("Can not locate the body for " &
                     Short_Source_Name (File_Find (El)));
               end if;

            end;

         else
            SF := File_Find (Completion_El);

            if not Present (SF) then
               SF := Add_Needed_Source
                 (To_String (Asis.Compilation_Units.Text_Name
                               (Asis.Elements.Enclosing_Compilation_Unit
                                  (Completion_El))));

               Gnatcheck.Diagnoses.Add_Line_To_Mapping_Table;
            end if;

         end if;

      end if;

   end Store_Needed_Source;

   -----------------------------
   -- Store_Task_Creation_Arc --
   -----------------------------

   procedure Store_Task_Creation_Arc (El : Asis.Element) is
      Arg_Kind  : constant Flat_Element_Kinds := Flat_Element_Kind (El);
      Task_Node :          GS_Node_Id;
   begin

      case Arg_Kind is
         when A_Single_Task_Declaration =>
            Task_Node := Corresponding_Task_Node (El, Single_Task => True);
            Store_Arc (Task_Node);
         when A_Variable_Declaration |
              A_Constant_Declaration =>

            declare
               Tasks            : constant Element_List := Names (El);
               Task_Def         :          Asis.Element;
               Task_Def_Node    :          GS_Node_Id;
               Task_Def_Defined :          Boolean      := False;
            begin

               for J in Tasks'Range loop
                  Task_Node := Find_Node (Tasks (J));

                  if No (Task_Node) then

                     if not Task_Def_Defined then
                        Task_Def         := Get_Root_Type (El);
                        Task_Def         := Type_Declaration_View (Task_Def);
                        Task_Def_Node    := Corresponding_Node (Task_Def);
                        Task_Def_Defined := True;
                     end if;

                     Task_Node      :=
                        Corresponding_Task_Node
                          (Tasks (J),
                           Single_Task   => False,
                           Task_Def_Node => Task_Def_Node);

                  end if;

                  Store_Arc (Task_Node);
               end loop;

            end;

         when others =>
            null;
            pragma Assert (False);
      end case;

   end Store_Task_Creation_Arc;

end Gnatcheck.Global_State;
