------------------------------------------------------------------------------
--                                                                          --
--                          GNATCHECK COMPONENTS                            --
--                                                                          --
--           G N A T C H E C K . R U L E S . R U L E _ T A B L E            --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2004-2005, AdaCore                     --
--                                                                          --
-- GNATCHECK  is  free  software;  you can redistribute it and/or modify it --
-- under terms of the  GNU  General Public License as published by the Free --
-- Software Foundation;  either version 2, or ( at your option)  any  later --
-- version.  GNATCHECK  is  distributed in the hope that it will be useful, --
-- but  WITHOUT  ANY  WARRANTY;   without  even  the  implied  warranty  of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details.  You should have received a copy of the --
-- GNU  General Public License distributed with GNAT; see file  COPYING. If --
-- not,  write to the  Free Software Foundation,  51 Franklin Street, Fifth --
-- Floor, Boston, MA 02110-1301, USA.                                       --
--                                                                          --
-- GNATCHECK is maintained by AdaCore (http://www.adacore.com).             --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Characters.Handling; use Ada.Characters.Handling;

with ASIS_UL.Output;          use ASIS_UL.Output;

with Gnatcheck.Compiler;      use Gnatcheck.Compiler;

package body Gnatcheck.Rules.Rule_Table is

   -----------------------
   -- Local subprograms --
   -----------------------

   procedure Turn_Rule_On (Param : String);
   pragma Unreferenced (Turn_Rule_On);
   --  Turns on the argument rule (casing is not important) If there is no
   --  rule with the given name, generates the corresponding warning message
   --  and does nothing

   procedure Turn_Rule_Off (Param : String);
   pragma Unreferenced (Turn_Rule_Off);
   --  Turns off the argument rule (casing is not important) If there is no
   --  rule with the given name, generates the corresponding warning message
   --  and does nothing

   function Get_Rule_Name (Param : String) return String;
   --  Returns the rule name from the parameter of +R and -R options

   --------------
   -- Get_Rule --
   --------------

   function Get_Rule (Rule_Name : String) return Rule_Id is
      Result               : Rule_Id          := No_Rule;
      Normalised_Rule_Name : constant String  := To_Lower (Rule_Name);
   begin

      --  This is a rather noneffective implementation. At some point we
      --  should think about a hash table and about more efficient rule
      --  names normalization

      for J in First_Rule .. All_Rules.Last loop

         if To_Lower (All_Rules.Table (J).Name.all) = Normalised_Rule_Name then
            Result := J;
            exit;
         end if;

      end loop;

      return Result;
   end Get_Rule;

   -------------------
   -- Get_Rule_Name --
   -------------------

   function Get_Rule_Name (Param : String) return String is
      First_Idx : constant Natural := Param'First;
      Last_Idx  :          Natural := Param'Last;
   begin

      for J in Param'Range loop
         if Param (J) = ':' then
            Last_Idx := J - 1;
            exit;
         end if;
      end loop;

      return Param (First_Idx .. Last_Idx);
   end Get_Rule_Name;

   -------------
   -- Present --
   -------------

   function Present (Rule : Rule_Id) return Boolean is
   begin
      return Rule in All_Rules.First .. All_Rules.Last;
   end Present;

   -------------------------
   -- Process_Rule_Option --
   -------------------------

   procedure Process_Rule_Option (Option : String) is
      First_Idx : constant Natural := Option'First;
      Last_Idx  : constant Natural := Option'Last;

      Word_Start : Natural := 0;
      Word_End   : Natural := 0;
      --  Should be set to select the next subword from Option - either rhe
      --  rule name or a rule parameter

      procedure Set_Parameter;
      --  Provided that Word_Start points to the beginning of the rule name or
      --  rule parameter, sets Word_Start and Word_Start to point to the next
      --  parameter, Sets Word_Start to 0 if there is no parameter any more.
      --  This procedure also checks the syntax of the rule oprion - that is,
      --  that the rule name is separated from parameter(s) by ':', and
      --  paremeters are separated by ',', if this check fails, Word_Start is
      --  set to 0.

      Rule    : Rule_Id;
      Enable  : Boolean;

      procedure Set_Parameter is
      begin
         if Word_End < Last_Idx then
            Word_Start := Word_End + 2;
         else
            Word_Start := 0;
         end if;

         if Word_Start > 0 then

            Word_End := Last_Idx;

            for J in Word_Start .. Last_Idx loop

               if Option (J) = ',' then
                  Word_End := J - 1;
                  exit;
               end if;

            end loop;

         end if;

      end Set_Parameter;

   begin

      if Option = "+ALL" then
         Turn_All_Rules_On;
         return;

      elsif Option = "+GLOBAL" then
         Turn_All_Global_Rules_On;
         return;

      elsif Option = "-ALL" then
         Turn_All_Rules_Off;
         return;

      elsif Option = "-GLOBAL" then
         Turn_All_Global_Rules_Off;
         return;
      end if;

      if Last_Idx - First_Idx > 2
       and then
         (Option (First_Idx) = '+'
         or else
          Option (First_Idx) = '-')
       and then
          Option (First_Idx + 1) = 'R'
      then
         Enable := Option (First_Idx) = '+';

         --  Computing the rule name and defining the rule

         Word_Start := First_Idx + 2;

         Word_End := Last_Idx;

         for J in Word_Start + 1 .. Last_Idx loop

            if Option (J) = ':' then
               Word_End := J - 1;
               exit;
            end if;

         end loop;

         --  Separate processing for restrictions, warnings and ordinary
         --  rules

         if To_Lower (Option (Word_Start .. Word_End)) = "restrictions" then

            Set_Parameter;

            if Word_Start = 0 then
               Error ("restrictions rule option must have a parameter");
               return;
            else

               if Enable then

                  while Word_Start /= 0 loop
                     Process_Restriction_Param
                       (Option (Word_Start .. Word_End),
                        Enable);

                     Set_Parameter;
                  end loop;

               else
                  Error ("-R option for restrictions is not implemented yet," &
                         "ignored");
                  return;
               end if;

            end if;

         elsif To_Lower (Option (Word_Start .. Word_End)) = "style_checks" then

            Set_Parameter;

            if Word_Start = 0 then
               Error ("style_checks rule option must have a parameter");
               return;
            else

               if Enable then

                  while Word_Start /= 0 loop
                     Process_Style_Check_Param
                       (Option (Word_Start .. Word_End),
                        Enable);

                     Set_Parameter;
                  end loop;

               else
                  Error ("-R option for style_checks is not implemented yet," &
                         "ignored");
                  return;
               end if;

            end if;

         elsif To_Lower (Option (Word_Start .. Word_End)) = "warnings" then

            Set_Parameter;

            if Word_Start = 0 then
               Error ("warnings rule option must have a parameter");
               return;
            else

               if Enable then

                  while Word_Start /= 0 loop
                     Process_Warning_Param
                       (Option (Word_Start .. Word_End),
                        Enable);

                     Set_Parameter;
                  end loop;

               else
                  Error ("-R option for warnings is not implemented yet," &
                         "ignored");
                  return;
               end if;

            end if;

         else

            Rule := Get_Rule (Option (Word_Start .. Word_End));

            if Present (Rule) then

               Set_Parameter;

               if Word_Start = 0 then

                  if Enable then
                     All_Rules.Table (Rule).Rule_State := Enabled;
                  else
                     All_Rules.Table (Rule).Rule_State := Disabled;
                  end if;

               else

                  while Word_Start /= 0 loop

                     Process_Rule_Parameter
                       (Rule    => All_Rules.Table (Rule).all,
                        Param   => Option (Word_Start .. Word_End),
                        Enable  => Enable);

                     Set_Parameter;
                  end loop;

               end if;

            else
               Error ("unknown rule : " & Option (Word_Start .. Word_End) &
                      " (ignored)");
            end if;

         end if;

      else
         Error ("unknown rule option: " & Option & ", ignored");
      end if;

   end Process_Rule_Option;

   ----------------
   -- Rules_Help --
   ----------------

   procedure Rules_Help is
   begin

      Info ("gnatcheck currently implements the following rules:");

      if All_Rules.Last < First_Rule then
         Info ("  There is no rule implemented");
      else

         for J in First_Rule .. All_Rules.Last loop
            Print_Rule_Help (All_Rules.Table (J).all);
         end loop;

      end if;

      Info ("gnatcheck allows to activate the following checks provided by " &
            "GNAT");
      Info ("using the same syntax to control these checks as for other " &
            "rules:");
      Info ("  Warnings     - compiler warnings");
      Info ("  Style_Checks - compiler style checks");
      Info ("  Restrictions - checks made by pragma Restriction_Warnings");

   end Rules_Help;

   ---------------------
   --  Set_Rule_State --
   ---------------------

   procedure Set_Rule_State (For_Rule : Rule_Id; To_State : Rule_States) is
   begin
      pragma Assert (Present (For_Rule));
      All_Rules.Table (For_Rule).Rule_State := To_State;
   end Set_Rule_State;

   ------------------------
   -- Turn_All_Rules_Off --
   ------------------------

   procedure Turn_All_Rules_Off is
   begin

      for J in All_Rules.First .. All_Rules.Last loop
         All_Rules.Table (J).Rule_State := Disabled;
      end loop;

   end Turn_All_Rules_Off;

   -------------------------------
   -- Turn_All_Global_Rules_Off --
   -------------------------------

   procedure Turn_All_Global_Rules_Off is
   begin

      for J in All_Rules.First .. All_Rules.Last loop

         if All_Rules.Table (J).all in Global_Rule_Template'Class then
            All_Rules.Table (J).Rule_State := Disabled;
         end if;

      end loop;

   end Turn_All_Global_Rules_Off;

   ------------------------------
   -- Turn_All_Global_Rules_On --
   ------------------------------

   procedure Turn_All_Global_Rules_On is
   begin

      for J in All_Rules.First .. All_Rules.Last loop

         if All_Rules.Table (J).all in Global_Rule_Template'Class then
            All_Rules.Table (J).Rule_State := Enabled;
         end if;

      end loop;

   end Turn_All_Global_Rules_On;

   -----------------------
   -- Turn_All_Rules_On --
   -----------------------

   procedure Turn_All_Rules_On is
   begin

      for J in All_Rules.First .. All_Rules.Last loop
         All_Rules.Table (J).Rule_State := Enabled;
      end loop;

   end Turn_All_Rules_On;

   -------------------
   -- Turn_Rule_Off --
   -------------------

   procedure Turn_Rule_Off (Param : String) is
      Rule_Name : constant String  := Get_Rule_Name (Param);
      Rule      : constant Rule_Id := Get_Rule (Rule_Name);
   begin

      if Present (Rule) then
         All_Rules.Table (Rule).Rule_State := Disabled;
      else
         Error ("unknown rule : " & Rule_Name & " (ignored)");
      end if;

   end Turn_Rule_Off;

   ------------------
   -- Turn_Rule_On --
   ------------------

   procedure Turn_Rule_On (Param : String) is
      Rule_Name : constant String  := Get_Rule_Name (Param);
      Rule      : constant Rule_Id := Get_Rule (Rule_Name);
   begin

      if Present (Rule) then
         All_Rules.Table (Rule).Rule_State := Enabled;
      else
         Error ("unknown rule : " & Rule_Name & " (ignored)");
      end if;

   end Turn_Rule_On;

end Gnatcheck.Rules.Rule_Table;
